# -*- coding: utf-8 -*-
from .annotator import PdfAnnotator
from .config.appearance import Appearance
from .config.location import Location
from .config.metadata import Metadata


__all__ = ['PdfAnnotator', 'Appearance', 'Location', 'Metadata']



#!/usr/bin/env vpython3
import sys
import time
import random
import io
import struct
import datetime
import hashlib
import codecs
import struct
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from endesive import signer
from endesive.pdf.PyPDF2 import pdf, generic as po


def EncodedString(s):
    return po.createStringObject(codecs.BOM_UTF16_BE + s.encode("utf-16be"))


class UnencryptedBytes(po.utils.bytes_type, po.PdfObject):
    original_bytes = property(lambda self: self)

    def writeToStream(self, stream, encryption_key):
        stream.write(b"<")
        stream.write(self)
        stream.write(b">")


class WNumberObject(po.NumberObject):
    Format = b"%08d"

    def writeToStream(self, stream, encryption_key):
        stream.write(self.Format % self)


class Main(pdf.PdfFileWriter):
    annottext = True
    annotbutton = True

    def encrypt(self, prev, password, rc):
        encrypt = prev.trailer["/Encrypt"].getObject()
        if encrypt["/V"] == 2:
            rev = 3
            keylen = int(128 / 8)
        else:
            rev = 2
            keylen = int(40 / 8)
        P = encrypt["/P"]
        O = encrypt["/O"]
        ID_1 = prev.trailer["/ID"][0]
        if rev == 2:
            U, key = pdf._alg34(password, O, P, ID_1)
        else:
            assert rev == 3
            U, key = pdf._alg35(password, rev, keylen, O, P, ID_1, False)
        self._encrypt_key = key

    def write(self, stream, prev, startdata):
        stream.write(pdf.b_("\r\n"))
        positions = {2: 0}
        for i in range(2, len(self._objects)):
            idnum = i + 1
            obj = self._objects[i]
            if obj is None:
                positions[idnum] = 0
                continue
            positions[idnum] = startdata + stream.tell()
            stream.write(pdf.b_(str(idnum) + " 0 obj\n"))
            key = None
            if self._encrypt_key is not None:
                pack1 = struct.pack("<i", i + 1)[:3]
                pack2 = struct.pack("<i", 0)[:2]
                key = self._encrypt_key + pack1 + pack2
                assert len(key) == (len(self._encrypt_key) + 5)
                md5_hash = hashlib.md5(key).digest()
                key = md5_hash[: min(16, len(self._encrypt_key) + 5)]
            obj.writeToStream(stream, key)
            stream.write(pdf.b_("\nendobj\n"))

        xref_location = startdata + stream.tell()
        if not prev.xrefstream:
            trailer = po.DictionaryObject()
        else:
            trailer = po.StreamObject()
            self._addObject(trailer)
        # xref table
        trailer.update(
            {
                po.NameObject("/Size"): po.NumberObject(len(self._objects) + 1),
                po.NameObject("/Root"): self.x_root,
                po.NameObject("/Info"): self.x_info,
                po.NameObject("/Prev"): po.NumberObject(prev.startxref),
                po.NameObject("/ID"): self._ID,
            }
        )
        if prev.isEncrypted:
            trailer[po.NameObject("/Encrypt")] = prev.trailer.raw_get("/Encrypt")
        if not prev.xrefstream:
            stream.write(pdf.b_("xref\n"))
            stream.write(pdf.b_("0 1\n"))
            stream.write(pdf.b_("0000000000 65535 f \n"))
            keys = sorted(positions.keys())
            i = 0
            while i < len(keys):
                off = positions[keys[i]]
                if off == 0:
                    while i < len(keys) and positions[keys[i]] == 0:
                        i += 1
                    start = i
                    while i < len(keys) and positions[keys[i]] != 0:
                        i += 1
                    stream.write(pdf.b_("%d %d \n" % (keys[start], i - start)))
                    i = start
                    continue
                else:
                    stream.write(pdf.b_("%010d %05d n \n" % (off, 0)))
                i += 1

            # trailer
            stream.write(pdf.b_("trailer\n"))
            trailer.writeToStream(stream, None)
        else:

            def pack(offset):
                return struct.pack(">q", offset)

            dataindex = ["0 1"]
            dataxref = [b"\x00" + pack(0)]
            keys = sorted(positions.keys())
            i = 0
            while i < len(keys):
                off = positions[keys[i]]
                if off != 0:
                    start = i
                    while i < len(keys) and positions[keys[i]] != 0:
                        dataxref.append(b"\x01" + pack(positions[keys[i]]))
                        i += 1
                    stop = i
                    dataindex.append("%d %d" % (keys[start], stop - start))
                else:
                    i += 1
            dataindex = " ".join(dataindex)
            dataxref = b"".join(dataxref)
            trailer[po.NameObject("/Type")] = po.NameObject("/XRef")
            trailer[po.NameObject("/W")] = po.NameObject("[1 8 0]")
            trailer[po.NameObject("/Index")] = po.NameObject("[%s]" % dataindex)
            trailer._data = dataxref
            retval = trailer.flateEncode()
            trailer.update(retval)
            trailer._data = retval._data
            stream.write(pdf.b_("%d 0 obj\n" % (len(self._objects))))
            trailer.writeToStream(stream, None)
            stream.write(pdf.b_("\nendobj"))

        # eof
        stream.write(pdf.b_("\nstartxref\n%s\n%%%%EOF\n" % (xref_location)))

    def _extend(self, obj):
        stream = getattr(obj, "stream", None)
        if stream is not None:
            d = {"__streamdata__": stream, "/Length": len(stream)}
            d.update(obj)
            dct = pdf.StreamObject.initializeFromDictionary(d)
        else:
            dct = pdf.DictionaryObject()
        for k, v in obj.items():
            if isinstance(v, pdf.DictionaryObject):
                if v.indirect:
                    v = self._extend(v)
                    v = self._addObject(v)
                else:
                    v = self._extend(v)
            elif isinstance(v, list):
                v = pdf.ArrayObject(v)
            dct[k] = v
        return dct

    def makepdf(self, prev, algomd, zeros):
        catalog = prev.trailer["/Root"]
        size = prev.trailer["/Size"]
        pages = catalog["/Pages"].getObject()
        page0ref = pages["/Kids"][0]

        while len(self._objects) < size - 1:
            self._objects.append(None)

        obj13 = po.DictionaryObject()
        obj13ref = self._addObject(obj13)
        obj12 = po.DictionaryObject()
        obj12ref = self._addObject(obj12)

        obj12.update(
            {
                po.NameObject("/Type"): po.NameObject("/Sig"),
                po.NameObject("/Filter"): po.NameObject("/Adobe.PPKLite"),
                po.NameObject("/SubFilter"): po.NameObject("/adbe.pkcs7.detached"),
                po.NameObject("/Name"): EncodedString("Example User"),
                po.NameObject("/Location"): EncodedString("Los Angeles, CA"),
                po.NameObject("/Reason"): EncodedString("Testing"),
                po.NameObject("/M"): EncodedString("D:20200317214832+01'00'"),
                po.NameObject("/Contents"): UnencryptedBytes(zeros),
                po.NameObject("/ByteRange"): po.ArrayObject(
                    [
                        WNumberObject(0),
                        WNumberObject(0),
                        WNumberObject(0),
                        WNumberObject(0),
                    ]
                ),
            }
        )
        obj13.update(
            {
                po.NameObject("/FT"): po.NameObject("/Sig"),
                po.NameObject("/Type"): po.NameObject("/Annot"),
                po.NameObject("/Subtype"): po.NameObject("/Widget"),
                po.NameObject("/F"): po.NumberObject(132),
                po.NameObject("/T"): EncodedString("Signature1"),
                po.NameObject("/V"): obj12ref,
                po.NameObject("/P"): page0ref,
                po.NameObject("/Rect"): po.ArrayObject(
                    [
                        po.FloatObject(0.0),
                        po.FloatObject(0.0),
                        po.FloatObject(0.0),
                        po.FloatObject(0.0),
                    ]
                ),
            }
        )

        if self.annottext:
            from endesive.pdf.PyPDF2_annotate.annotations.text import FreeText
            from endesive.pdf.PyPDF2_annotate.annotations.image import Image
            from endesive.pdf.PyPDF2_annotate.config.appearance import Appearance
            from endesive.pdf.PyPDF2_annotate.config.location import Location
            from endesive.pdf.PyPDF2_annotate.util.geometry import identity

            annotationtext = None
            #annotationtext = "User signature text"
            x1, y1, x2, y2 = (470, 0, 570, 100)
            if annotationtext is not None:
                annotation = FreeText(
                    Location(x1=x1, y1=y1, x2=x2, y2=y2, page=0),
                    Appearance(
                        fill=[0, 0, 0],
                        stroke_width=1,
                        wrap_text=True,
                        font_size=12,
                        content=annotationtext,
                    ),
                )
                names = ("BS", "C", "Contents", "DA")
                if not self.annotbutton:
                    obj13[po.NameObject("/Subtype")] = po.NameObject("/FreeText")
            else:
                from PIL import Image as PILImage

                image = PILImage.open("signature_test.png")
                ap = Appearance()
                ap.image = image
                annotation = Image(Location(x1=x1, y1=y1, x2=x2, y2=y2, page=0), ap)
                if not self.annotbutton:
                    names = (
                        #
                        "Subtype",
                    )
                else:
                    names = ()

            pdfa = annotation.as_pdf_object(identity(), page=page0ref)
            objapn = self._extend(pdfa["/AP"]["/N"])
            objapnref = self._addObject(objapn)

            for name in names + (
                "Rect",
                # "Subtype",
            ):
                key = po.NameObject("/" + name)
                v = pdfa[key]
                obj13[key] = v

            objap = po.DictionaryObject()
            objap[po.NameObject("/N")] = objapnref
            obj13.update(
                {
                    po.NameObject("/AP"): objap,
                    po.NameObject("/SM"): po.createStringObject("TabletPOSinline"),
                }
            )

            page0 = page0ref.getObject()
            annots = po.ArrayObject([obj13ref])
            if "/Annots" in page0:
                page0annots = page0["/Annots"]
                if isinstance(page0annots, po.IndirectObject):
                    annots.insert(0, page0annots)
                elif isinstance(page0annots, po.ArrayObject):
                    annots = page0annots
                    annots.append(obj13ref)
            page0.update({po.NameObject("/Annots"): annots})
            self._objects[page0ref.idnum - 1] = page0

        if "/Perms" not in catalog:
            obj10 = po.DictionaryObject()
            obj10ref = self._addObject(obj10)
            obj11 = po.DictionaryObject()
            obj11ref = self._addObject(obj11)
            obj14 = po.DictionaryObject()
            obj14ref = self._addObject(obj14)
            obj14.update({po.NameObject("/DocMDP"): obj12ref})
            obj10.update(
                {
                    po.NameObject("/Type"): po.NameObject("/TransformParams"),
                    po.NameObject("/P"): po.NumberObject(2),
                    po.NameObject("/V"): po.NameObject("/1.2"),
                }
            )
            obj11.update(
                {
                    po.NameObject("/Type"): po.NameObject("/SigRef"),
                    po.NameObject("/TransformMethod"): po.NameObject("/DocMDP"),
                    po.NameObject("/DigestMethod"): po.NameObject("/" + algomd.upper()),
                    po.NameObject("/TransformParams"): obj10ref,
                }
            )
            obj12[po.NameObject("/Reference")] = po.ArrayObject([obj11ref])
            catalog[po.NameObject("/Perms")] = obj14ref

        if "/AcroForm" in catalog:
            form = catalog["/AcroForm"].getObject()
            if "/Fields" in form:
                fields = form["/Fields"]
            else:
                fields = po.ArrayObject()
            fields.append(obj13ref)
            form.update(
                {
                    po.NameObject("/Fields"): fields,
                    po.NameObject("/SigFlags"): po.NumberObject(3),
                }
            )
            formref = catalog.raw_get("/AcroForm")
            if isinstance(formref, po.IndirectObject):
                self._objects[formref.idnum - 1] = form
                form = formref
        else:
            form = po.DictionaryObject()
            form.update(
                {
                    po.NameObject("/Fields"): po.ArrayObject([obj13ref]),
                    po.NameObject("/SigFlags"): po.NumberObject(3),
                }
            )
        catalog[po.NameObject("/AcroForm")] = form

        if "/Metadata" in catalog:
            catalog[po.NameObject("/Metadata")] = catalog.raw_get("/Metadata")

        x_root = prev.trailer.raw_get("/Root")
        self._objects[x_root.idnum - 1] = catalog
        self.x_root = po.IndirectObject(x_root.idnum, 0, self)
        self.x_info = prev.trailer.raw_get("/Info")

    def sign(self, md, algomd):
        tspurl = "http://public-qlts.certum.pl/qts-17"
        tspurl = None
        with open("ca/demo2_user1.p12", "rb") as fp:
            p12 = pkcs12.load_key_and_certificates(
                fp.read(), b"1234", backends.default_backend()
            )
        contents = signer.sign(
            None, p12[0], p12[1], p12[2], algomd, True, md, None, False, tspurl
        )
        return contents

    def main(self, fname, password):
        with open(fname, "rb") as fi:
            datau = fi.read()
        startdata = len(datau)

        fi = io.BytesIO(datau)

        prev = pdf.PdfFileReader(fi)
        if prev.isEncrypted:
            rc = prev.decrypt(password)
        else:
            rc = 0

        algomd = "sha1"
        aligned = False

        obj = prev.trailer
        for k in ("/Root", "/Perms", "/DocMDP", "/Reference"):
            if k in obj:
                obj = obj[k]
                if isinstance(obj, po.ArrayObject):
                    obj = obj[0]
                obj = obj.getObject()
            else:
                obj = None
                break
        if obj is not None:
            algomd = obj["/DigestMethod"][1:].lower()

        if aligned:
            zeros = b"0" * 37888
        else:
            md = getattr(hashlib, algomd)().digest()
            contents = self.sign(md, algomd)
            zeros = contents.hex().encode("utf-8")

        self.makepdf(prev, algomd, zeros)

        if prev.isEncrypted:
            self.encrypt(prev, password, rc)
        else:
            self._encrypt_key = None
        ID = prev.trailer.get("/ID", None)
        if ID is None:
            ID = po.ByteStringObject(hashlib.md5(repr(time.time()).encode()).digest())
        else:
            ID = ID[0]
        self._ID = po.ArrayObject(
            [
                ID,
                po.ByteStringObject(
                    hashlib.md5(repr(random.random()).encode()).digest()
                ),
            ]
        )

        fo = io.BytesIO()
        self.write(fo, prev, startdata)
        datas = fo.getvalue()

        br = [0, 0, 0, 0]
        bfrom = (b"[ " + b" ".join([WNumberObject.Format] * 4) + b" ]") % tuple(br)

        pdfbr1 = datas.find(zeros)
        pdfbr2 = pdfbr1 + len(zeros)
        br = [
            0,
            startdata + pdfbr1 - 1,
            startdata + pdfbr2 + 1,
            len(datas) - pdfbr2 - 1,
        ]
        bto = b"[%d %d %d %d]" % tuple(br)
        bto += b" " * (len(bfrom) - len(bto))
        assert len(bfrom) == len(bto)
        datas = datas.replace(bfrom, bto, 1)

        md = getattr(hashlib, algomd)()
        md.update(datau)
        b1 = datas[: br[1] - startdata]
        b2 = datas[br[2] - startdata :]
        md.update(b1)
        md.update(b2)
        md = md.digest()

        contents = self.sign(md, algomd)
        contents = contents.hex().encode("utf-8")
        if aligned:
            nb = len(zeros) - len(contents)
            contents += b"0" * nb
        datas = datas.replace(zeros, contents, 1)

        fname = fname.replace(".pdf", "-signed-pypdf.pdf")
        with open(fname, "wb") as fp:
            fp.write(datau)
            fp.write(datas)


def main():
    if len(sys.argv) > 1:
        cls = Main()
        cls.main(sys.argv[1], sys.argv[2] if len(sys.argv) > 2 else "")
    else:
        cls = Main()
        cls.main("pdf.pdf", "")

        cls = Main()
        cls.main("pdf-encrypted.pdf", "1234")


main()



#!/usr/bin/env vpython3
# coding: utf-8

import os
import sys
import sysconfig

os.environ['SOFTHSM2_CONF'] = 'softhsm2.conf'
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2.conf')):
    open('softhsm2.conf', 'wt').write('''\
log.level = DEBUG
directories.tokendir = %s/softhsm2/
objectstore.backend = file
slots.removable = false
''' % os.getcwd())
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2')):
    os.mkdir(os.path.join(os.getcwd(), 'softhsm2'))

#
#!/bin/bash
#SOFTHSM2_CONF=softhsm2.conf
#softhsm2-util --label "endesive" --slot 1 --init-token --pin secret1 --so-pin secret2
#softhsm2-util --show-slots
#

if sys.platform == 'win32':
    dllpath = r'W:\binw\SoftHSM2\lib\softhsm2-x64.dll'
else:
    dllpath = os.path.join(sysconfig.get_config_var('LIBDIR'), "softhsm/libsofthsm2.so")

from endesive import hsm
import PyKCS11 as PK11

'''
Create two certificates:
1. self signed hsm CA certificate with serial equal to HSM keyID=0x01
2. hsm USER 1 certificate with serial equal to HSM keyID=0x666690
'''
class HSM(hsm.HSM):
    def main(self):
        cakeyID = bytes((0x1,))
        rec = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, cakeyID)])
        if len(rec) == 0:
            label = 'hsm CA'
            self.gen_privkey(label, cakeyID)
            self.ca_gen(label, cakeyID, 'hsm CA')

        keyID = bytes((0x66,0x66,0x90))
        rec = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, keyID)])
        if len(rec) == 0:
            label = 'hsm USER 1'
            self.gen_privkey(label, keyID)
            self.ca_sign(keyID, label, 0x666690, "hsm USER 1", 365, cakeyID)

        self.cert_export('cert-hsm-ca', cakeyID)
        self.cert_export('cert-hsm-user1', keyID)

def main():
    cls = HSM(dllpath)
    cls.create("endesieve", "secret1", "secret2")
    cls.login("endesieve", "secret1")
    try:
        cls.main()
    finally:
        cls.logout()
main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from asn1crypto import (
    algos,
    cms,
    core,
    crl,
    keys,
    ocsp,
    parser,
    pdf,
    pem,
    pkcs12,
    tsp,
    util,
    version,
    x509,
)
from endesive.pdf.PyPDF2 import pdf, generic as po
from endesive.signer import cert2asn


class Dumper:
    def __init__(self, fname):
        self.pdf = pdf.PdfFileReader(open(fname, "rb"))

    def show(self):
        root = self.pdf.trailer["/Root"]
        dss = root["/DSS"]
        print(dss)
        self.showCerts(dss["/Certs"])
        self.showVRI(dss["/VRI"])
        self.showOCSPs(dss["/OCSPs"])
        self.showCRLs(dss["/CRLs"])

    def showCert(self, der):
        pem = cert2asn(der, False)
        # pem.debug()
        print(pem.serial_number)
        print(pem.issuer.native)
        print(pem.subject.native)

    def showOCSP(self, der):
        data = ocsp.OCSPResponse.load(der)
        print(data.native["response_status"])
        certs = data.basic_ocsp_response.native["certs"]
        cert = certs[0]["tbs_certificate"]
        print(cert["serial_number"])
        print(cert["subject"])
        # data.debug()

    def showCerts(self, objs):
        print("*" * 20, "Certs")
        for robj in objs:
            obj = robj.getObject()
            der = obj.getData()
            self.showCert(der)

    def showVRI(self, obj):
        print("*" * 20, "VRI")
        for k, robj in obj.items():
            print("*" * 20, k)
            vobj = robj.getObject()
            print("/TU", vobj["/TU"])
            print("*" * 10, "VRI-cert")
            for robj in vobj["/Cert"]:
                der = robj.getObject().getData()
                self.showCert(der)
            try:
                robjs = vobj["/OCSP"]
            except KeyError:
                robjs = []
            for robj in robjs:
                print("*" * 10, "VRI-OCSP")
                der = robj.getObject().getData()
                self.showOCSP(der)

    def showOCSPs(self, objs):
        print("*" * 20, "OCSPs")
        for robj in objs:
            obj = robj.getObject()
            der = obj.getData()
            print("*" * 10, "OCSP")
            self.showOCSP(der)

    def showCRLs(self, objs):
        print("*" * 20, "CRLs")
        for robj in objs:
            obj = robj.getObject()
            print(obj)


def main():
    cls = Dumper("pdf-signed-cms-m32_ocsp.pdf")
    cls.show()


main()



# *-* coding: utf-8 *-*
import sys
from email import message_from_string

from asn1crypto import cms
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.asymmetric import padding


class DecryptedData(object):

    def decrypt(self, data, key):
        msg = message_from_string(data)
        data = None
        for part in msg.walk():
            # multipart/* are just containers
            if part.get_content_maintype() == 'multipart':
                continue
            if part.get_content_type() not in (
                'application/x-pkcs7-mime',
                'application/pkcs7-mime',
            ):
                continue
            data = part.get_payload(decode=True)
            break
        if data is None:
            return

        signed_data = cms.ContentInfo.load(data)['content']
        # signed_data.debug()

        algo = signed_data['encrypted_content_info']['content_encryption_algorithm']['algorithm'].native
        param = signed_data['encrypted_content_info']['content_encryption_algorithm']['parameters'].native
        edata = signed_data['encrypted_content_info']['encrypted_content'].native
        pkey = signed_data['recipient_infos'].native[0]['encrypted_key']

        keyalgo = signed_data['recipient_infos'].native[0]['key_encryption_algorithm']
        if keyalgo['algorithm'] == 'rsaes_oaep':
            keyparam = keyalgo['parameters']
            mga = keyparam['mask_gen_algorithm']
            mgh = getattr(hashes, mga['parameters']['algorithm'].upper())()
            pad = padding.OAEP(
                mgf=getattr(padding, mga['algorithm'].upper())(algorithm=mgh),
                algorithm=getattr(hashes, keyparam['hash_algorithm']['algorithm'].upper())(),
                label=keyparam['p_source_algorithm']['parameters']
            )
            udata = key.decrypt(pkey, pad)
        elif keyalgo['algorithm'] == 'rsaes_pkcs1v15':
            udata = key.decrypt(pkey, padding.PKCS1v15())
        else:
            raise ValueError('Unknown key algorithm', keyalgo['algorithm'])

        algorithm, mode = algo.split('_', 1)
        algorithm = algorithm.upper()
        if algorithm in (
            'AES128',
            'AES192',
            'AES256',
        ):
            cipher = Cipher(
                algorithms.AES(udata),
                getattr(modes, mode.upper())(param),
                default_backend()
            )
        elif algorithm == 'TRIPLEDES':
            # XXX will be removed in version 48.0.0
            from cryptography.hazmat.decrepit.ciphers.algorithms import TripleDES
            # XXX howto decode parameters to CBC mode ?
            mode = 'cbc'
            cipher = Cipher(
                TripleDES(udata),
                getattr(modes, mode.upper())(param),
                default_backend()
            )
        else:
            raise ValueError('Unknown algorithm', algo)

        decryptor = cipher.decryptor()
        udata = decryptor.update(edata) + decryptor.finalize()
        #if keyalgo['algorithm'] != 'rsaes_oaep':
        nb = ord(udata[-1]) if sys.version[0] < '3' else udata[-1]
        udata = udata[:-nb]
        return udata


def decrypt(data, key):
    cls = DecryptedData()
    return cls.decrypt(data, key)



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
from PyPDF2 import PdfFileReader, PdfFileWriter

fname = "pdf.pdf"
with open(fname, "rb") as in_file:
    input_pdf = PdfFileReader(in_file)

    output_pdf = PdfFileWriter()
    output_pdf.appendPagesFromReader(input_pdf)
    output_pdf.encrypt("1234", "1234")

    fname = fname.replace('.pdf', '-encrypted.pdf')
    with open(fname, "wb") as out_file:
        output_pdf.write(out_file)



# -*- coding: iso-8859-1 -*-

"PDF Template Helper for FPDF.py"

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010 Mariano Reingart"
__license__ = "LGPL 3.0"

import sys,os,csv
from .fpdf import FPDF
from .py3k import PY3K, basestring, unicode

def rgb(col):
    return (col // 65536), (col // 256 % 256), (col% 256)

class Template:
    def __init__(self, infile=None, elements=None, format='A4', orientation='portrait',
                 title='', author='', subject='', creator='', keywords=''):
        if elements:
            self.load_elements(elements)
        self.handlers = {'T': self.text, 'L': self.line, 'I': self.image, 
                         'B': self.rect, 'BC': self.barcode, 'W': self.write, }
        self.texts = {}
        pdf = self.pdf = FPDF(format=format,orientation=orientation, unit="mm")
        pdf.set_title(title)
        pdf.set_author(author)
        pdf.set_creator(creator)
        pdf.set_subject(subject)
        pdf.set_keywords(keywords)

    def load_elements(self, elements):
        "Initialize the internal element structures"
        self.pg_no = 0
        self.elements = elements
        self.keys = [v['name'].lower() for v in self.elements]
    
    def parse_csv(self, infile, delimiter=",", decimal_sep="."):
        "Parse template format csv file and create elements dict"
        keys = ('name','type','x1','y1','x2','y2','font','size',
            'bold','italic','underline','foreground','background',
            'align','text','priority', 'multiline')
        self.elements = []
        self.pg_no = 0
        if not PY3K:
            f = open(infile, 'rb')
        else:
            f = open(infile)
        for row in csv.reader(f, delimiter=delimiter):
            kargs = {}
            for i,v in enumerate(row):
                if not v.startswith("'") and decimal_sep!=".": 
                    v = v.replace(decimal_sep,".")
                else:
                    v = v
                if v=='':
                    v = None
                else:
                    v = eval(v.strip())
                kargs[keys[i]] = v
            self.elements.append(kargs)
        self.keys = [v['name'].lower() for v in self.elements]

    def add_page(self):
        self.pg_no += 1
        self.texts[self.pg_no] = {}
        
    def __setitem__(self, name, value):
        if name.lower() in self.keys:
            if not PY3K and isinstance(value, unicode):
                value = value.encode("latin1","ignore")
            elif value is None:
                value = ""
            else:
                value = str(value)
            self.texts[self.pg_no][name.lower()] = value

    # setitem shortcut (may be further extended)
    set = __setitem__

    def has_key(self, name):
        return name.lower() in self.keys
        
    def __getitem__(self, name):
        if name in self.keys:
            key = name.lower()
            if key in self.texts:
                # text for this page:
                return self.texts[self.pg_no][key]
            else:
                # find first element for default text:
                elements = [element for element in self.elements
                    if element['name'].lower() == key]
                if elements:
                    return elements[0]['text']

    def split_multicell(self, text, element_name):
        "Divide (\n) a string using a given element width"
        pdf = self.pdf
        element = [element for element in self.elements
            if element['name'].lower() == element_name.lower()][0]
        style = ""
        if element['bold']: style += "B"
        if element['italic']: style += "I"
        if element['underline']: style += "U"
        pdf.set_font(element['font'],style,element['size'])
        align = {'L':'L','R':'R','I':'L','D':'R','C':'C','':''}.get(element['align']) # D/I in spanish
        if isinstance(text, unicode) and not PY3K:
            text = text.encode("latin1","ignore")
        else:
            text = str(text)
        return pdf.multi_cell(w=element['x2']-element['x1'],
                             h=element['y2']-element['y1'],
                             txt=text,align=align,split_only=True)
        
    def render(self, outfile, dest="F"):
        pdf = self.pdf
        for pg in range(1, self.pg_no+1):
            pdf.add_page()
            pdf.set_font('Arial','B',16)
            pdf.set_auto_page_break(False,margin=0)

            for element in sorted(self.elements,key=lambda x: x['priority']):
                #print "dib",element['type'], element['name'], element['x1'], element['y1'], element['x2'], element['y2']
                element = element.copy()
                element['text'] = self.texts[pg].get(element['name'].lower(), element['text'])
                if 'rotate' in element:
                    pdf.rotate(element['rotate'], element['x1'], element['y1'])
                self.handlers[element['type'].upper()](pdf, **element)
                if 'rotate' in element:
                    pdf.rotate(0)
        
        if dest:
            return pdf.output(outfile, dest)
        
    def text(self, pdf, x1=0, y1=0, x2=0, y2=0, text='', font="arial", size=10, 
             bold=False, italic=False, underline=False, align="", 
             foreground=0, backgroud=65535, multiline=None,
             *args, **kwargs):
        if text:
            if pdf.text_color!=rgb(foreground):
                pdf.set_text_color(*rgb(foreground))
            if pdf.fill_color!=rgb(backgroud):
                pdf.set_fill_color(*rgb(backgroud))

            font = font.strip().lower()
            if font == 'arial black':
                font = 'arial'
            style = ""
            for tag in 'B', 'I', 'U':
                if (text.startswith("<%s>" % tag) and text.endswith("</%s>" %tag)):
                    text = text[3:-4]
                    style += tag
            if bold: style += "B"
            if italic: style += "I"
            if underline: style += "U"
            align = {'L':'L','R':'R','I':'L','D':'R','C':'C','':''}.get(align) # D/I in spanish
            pdf.set_font(font,style,size)
            ##m_k = 72 / 2.54
            ##h = (size/m_k)
            pdf.set_xy(x1,y1)
            if multiline is None:
                # multiline==None: write without wrapping/trimming (default)
                pdf.cell(w=x2-x1,h=y2-y1,txt=text,border=0,ln=0,align=align)
            elif multiline:
                # multiline==True: automatic word - warp
                pdf.multi_cell(w=x2-x1,h=y2-y1,txt=text,border=0,align=align)
            else:
                # multiline==False: trim to fit exactly the space defined
                text = pdf.multi_cell(w=x2-x1, h=y2-y1,
                             txt=text, align=align, split_only=True)[0]
                print("trimming: *%s*" % text)
                pdf.cell(w=x2-x1,h=y2-y1,txt=text,border=0,ln=0,align=align)

            #pdf.Text(x=x1,y=y1,txt=text)

    def line(self, pdf, x1=0, y1=0, x2=0, y2=0, size=0, foreground=0, *args, **kwargs):
        if pdf.draw_color!=rgb(foreground):
            #print "SetDrawColor", hex(foreground)
            pdf.set_draw_color(*rgb(foreground))
        #print "SetLineWidth", size
        pdf.set_line_width(size)
        pdf.line(x1, y1, x2, y2)

    def rect(self, pdf, x1=0, y1=0, x2=0, y2=0, size=0, foreground=0, backgroud=65535, *args, **kwargs):
        if pdf.draw_color!=rgb(foreground):
            pdf.set_draw_color(*rgb(foreground))
        if pdf.fill_color!=rgb(backgroud):
            pdf.set_fill_color(*rgb(backgroud))
        pdf.set_line_width(size)
        pdf.rect(x1, y1, x2-x1, y2-y1)

    def image(self, pdf, x1=0, y1=0, x2=0, y2=0, text='', *args,**kwargs):
        if text:
            pdf.image(text,x1,y1,w=x2-x1,h=y2-y1,type='',link='')

    def barcode(self, pdf, x1=0, y1=0, x2=0, y2=0, text='', font="arial", size=1,
             foreground=0, *args, **kwargs):
        if pdf.draw_color!=rgb(foreground):
            pdf.set_draw_color(*rgb(foreground))
        font = font.lower().strip()
        if font == 'interleaved 2of5 nt':
            pdf.interleaved2of5(text,x1,y1,w=size,h=y2-y1)

    # Added by Derek Schwalenberg Schwalenberg1013@gmail.com to allow (url) links in templates (using write method) 2014-02-22
    def write(self, pdf, x1=0, y1=0, x2=0, y2=0, text='', font="arial", size=1,
              bold=False, italic=False, underline=False, align="", link='http://example.com',
             foreground=0, *args, **kwargs):
        if pdf.text_color!=rgb(foreground):
            pdf.set_text_color(*rgb(foreground))
        font = font.strip().lower()
        if font == 'arial black':
            font = 'arial'
        style = ""
        for tag in 'B', 'I', 'U':
            if (text.startswith("<%s>" % tag) and text.endswith("</%s>" %tag)):
                text = text[3:-4]
                style += tag
        if bold: style += "B"
        if italic: style += "I"
        if underline: style += "U"
        align = {'L':'L','R':'R','I':'L','D':'R','C':'C','':''}.get(align) # D/I in spanish
        pdf.set_font(font,style,size)
        ##m_k = 72 / 2.54
        ##h = (size/m_k)
        pdf.set_xy(x1,y1)
        pdf.write(5,text,link)



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from oscrypto import keys
from endesive import email


def main():
    with open('ca/demo2_user1.p12', 'rb') as f:
        key, cert, certe = keys.parse_pkcs12(f.read(), b'1234')
    datau = open('smime-unsigned.txt', 'rb').read()
    datas = email.sign(datau,
        key, cert, certe,
        'sha512',
        attrs=True,
        pss=True
    )
    open('smime-signed-pss-oscrypto.txt', 'wb').write(datas)


main()



#!/usr/bin/env vpython3
import sys
import time
import random
import io
import struct
import datetime
import hashlib
import codecs
import struct
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.hazmat.primitives.asymmetric import ec
from endesive import signer
from endesive.pdf.PyPDF2 import pdf, generic as po


def EncodedString(s):
    return po.createStringObject(codecs.BOM_UTF16_BE + s.encode("utf-16be"))


class UnencryptedBytes(po.utils.bytes_type, po.PdfObject):
    original_bytes = property(lambda self: self)

    def writeToStream(self, stream, encryption_key):
        stream.write(b"<")
        stream.write(self)
        stream.write(b">")


class WNumberObject(po.NumberObject):
    Format = b"%08d"

    def writeToStream(self, stream, encryption_key):
        stream.write(self.Format % self)


class SignedData(pdf.PdfFileWriter):
    def encrypt(self, prev, password, rc):
        encrypt = prev.trailer["/Encrypt"].getObject()
        if encrypt["/V"] == 2:
            rev = 3
            keylen = 128 // 8
        else:
            rev = 2
            keylen = 40 // 8
        P = encrypt["/P"].getObject()
        O = encrypt["/O"].getObject()
        ID_1 = prev.trailer["/ID"].getObject()[0]
        real_U = encrypt["/U"].getObject().original_bytes
        if rev == 2:
            U, key = pdf._alg34(password, O, P, ID_1)
        else:
            assert rev == 3
            U, key = pdf._alg35(
                password,
                rev,
                keylen,
                O,
                P,
                ID_1,
                encrypt.get("/EncryptMetadata", pdf.BooleanObject(False)).getObject(),
            )
            U, real_U = U[:16], real_U[:16]
        assert U == real_U
        self._encrypt_key = key

    def write(self, stream, prev, startdata):
        stream.write(pdf.b_("\r\n"))
        positions = {}
        for i in range(len(self._objects)):
            idnum = i + 1
            obj = self._objects[i]
            if obj is None:
                positions[idnum] = 0
                continue
            positions[idnum] = startdata + stream.tell()
            stream.write(pdf.b_(str(idnum) + " 0 obj\n"))
            key = None
            if self._encrypt_key is not None:
                pack1 = struct.pack("<i", i + 1)[:3]
                pack2 = struct.pack("<i", 0)[:2]
                key = self._encrypt_key + pack1 + pack2
                assert len(key) == (len(self._encrypt_key) + 5)
                md5_hash = hashlib.md5(key).digest()
                key = md5_hash[: min(16, len(self._encrypt_key) + 5)]
            obj.writeToStream(stream, key)
            stream.write(pdf.b_("\nendobj\n"))

        xref_location = startdata + stream.tell()
        if not prev.xrefstream:
            trailer = po.DictionaryObject()
        else:
            trailer = po.StreamObject()
            self._addObject(trailer)
        # xref table
        trailer.update(
            {
                po.NameObject("/Size"): po.NumberObject(len(self._objects) + 1),
                po.NameObject("/Root"): self.x_root,
                po.NameObject("/Prev"): po.NumberObject(prev.startxref),
                po.NameObject("/ID"): self._ID,
            }
        )

        if prev.isEncrypted:
            trailer[po.NameObject("/Encrypt")] = prev.trailer.raw_get("/Encrypt")

        if self.x_info:
            trailer[po.NameObject("/Info")] = self.x_info

        if not prev.xrefstream:
            stream.write(pdf.b_("xref\n"))
            positions[0] = 1
            keys = sorted(positions.keys())
            i = 0
            while i < len(keys):
                start = i
                while i < len(keys) and positions[keys[i]] != 0:
                    i += 1
                stream.write(pdf.b_("%d %d \n" % (keys[start], i - start)))
                i = start
                while i < len(keys) and positions[keys[i]] != 0:
                    if i == 0:
                        stream.write(pdf.b_("0000000000 65535 f \n"))
                    else:
                        stream.write(
                            pdf.b_("%010d %05d n \n" % (positions[keys[i]], 0))
                        )
                    i += 1
                while i < len(keys) and positions[keys[i]] == 0:
                    i += 1

            # trailer
            stream.write(pdf.b_("trailer\n"))
            trailer.writeToStream(stream, None)
        else:

            def pack(offset):
                return struct.pack(">q", offset)

            dataindex = ["0 1"]
            dataxref = [b"\x00" + pack(0)]
            keys = sorted(positions.keys())
            i = 0
            while i < len(keys):
                off = positions[keys[i]]
                if off != 0:
                    start = i
                    while i < len(keys) and positions[keys[i]] != 0:
                        dataxref.append(b"\x01" + pack(positions[keys[i]]))
                        i += 1
                    dataindex.append("%d %d" % (keys[start], i - start))
                else:
                    i += 1
            dataindex = " ".join(dataindex)
            dataxref = b"".join(dataxref)
            trailer[po.NameObject("/Type")] = po.NameObject("/XRef")
            trailer[po.NameObject("/W")] = po.NameObject("[1 8 0]")
            trailer[po.NameObject("/Index")] = po.NameObject("[%s]" % dataindex)
            trailer._data = dataxref
            retval = trailer.flateEncode()
            trailer.update(retval)
            trailer._data = retval._data
            stream.write(pdf.b_("%d 0 obj\n" % (len(self._objects))))
            trailer.writeToStream(stream, None)
            stream.write(pdf.b_("\nendobj"))

        # eof
        stream.write(pdf.b_("\nstartxref\n%s\n%%%%EOF\n" % (xref_location)))

    def _extend(self, obj):
        stream = getattr(obj, "stream", None)
        if stream is not None:
            # stream = stream.encode("utf-16be")
            d = {"__streamdata__": stream, "/Length": len(stream)}
            d.update(obj)
            dct = pdf.StreamObject.initializeFromDictionary(d)
            if "/Filter" in obj and obj["/Filter"] == "/FlatDecode":
                del dct["/Filter"]
                dct = dct.flateEncode()
        else:
            dct = pdf.DictionaryObject()
        for k, v in obj.items():
            if isinstance(v, pdf.DictionaryObject):
                if v.indirect:
                    v = self._extend(v)
                    v = self._addObject(v)
                else:
                    v = self._extend(v)
            elif isinstance(v, list):
                result = pdf.ArrayObject()
                for va in v:
                    if isinstance(va, pdf.DictionaryObject):
                        if va.indirect:
                            va = self._extend(va)
                            va = self._addObject(va)
                        else:
                            va = self._extend(va)
                    result.append(va)
                v = result
            dct[k] = v
        return dct

    def _make_signature(self, Contents=None, Type=None, SubFilter=None):
        sig = po.DictionaryObject()
        sig_ref = self._addObject(sig)

        sig.update(
            {
                po.NameObject("/Type"): Type,
                po.NameObject("/Filter"): po.NameObject("/Adobe.PPKLite"),
                po.NameObject("/SubFilter"): SubFilter,
                po.NameObject("/ByteRange"): po.ArrayObject(
                    [
                        WNumberObject(0),
                        WNumberObject(0),
                        WNumberObject(0),
                        WNumberObject(0),
                    ]
                ),
                po.NameObject("/Contents"): Contents,
            }
        )
        return sig, sig_ref

    def _make_sig_annotation(self, F=None, Vref=None, T=None, Pref=None):
        annot = po.DictionaryObject()
        annot_ref = self._addObject(annot)
        annot.update(
            {
                po.NameObject("/FT"): po.NameObject("/Sig"),
                po.NameObject("/Type"): po.NameObject("/Annot"),
                po.NameObject("/Subtype"): po.NameObject("/Widget"),
                po.NameObject("/F"): F,
                po.NameObject("/T"): T,
                po.NameObject("/V"): Vref,
                po.NameObject("/P"): Pref,
                # For an invisible signature, /Rect should be a size 0 box
                # Defaulting to that
                po.NameObject("/Rect"): po.ArrayObject(
                    [
                        po.FloatObject(0.0),
                        po.FloatObject(0.0),
                        po.FloatObject(0.0),
                        po.FloatObject(0.0),
                    ]
                ),
            }
        )
        return annot, annot_ref

    def addAnnotation(self, cert, udct, box, page0ref, obj13, obj13ref, new_13):
        from endesive.pdf.PyPDF2_annotate.annotations.signature import Signature
        from endesive.pdf.PyPDF2_annotate.config.appearance import Appearance
        from endesive.pdf.PyPDF2_annotate.config.location import Location
        from endesive.pdf.PyPDF2_annotate.util.geometry import identity

        x1, y1, x2, y2 = box
        annotation = Signature(
            Location(x1=x1, y1=y1, x2=x2, y2=y2, page=0), Appearance()
        )
        if "signature" in udct:
            # Plain text signature with the default font
            # text to render is contained in udct['signature']
            # font parameters are in udct['signature']['text']
            annotationtext = udct["signature"]
            wrap_text = udct.get("text", {}).get("wraptext", True)
            font_size = udct.get("text", {}).get("fontsize", 12)
            text_align = udct.get("text", {}).get("textalign", "left")
            line_spacing = udct.get("text", {}).get("linespacing", 1.2)

            annotation.add_default_font()
            annotation.set_signature_appearance(
                ["fill_colour", 0, 0, 0],
                ["font", "default", font_size],
                [
                    "text_box",
                    annotationtext,
                    "default",
                    0,
                    0,
                    x2 - x1,
                    y2 - y1,
                    font_size,
                    wrap_text,
                    text_align,
                    "middle",
                    line_spacing,
                ],
            )
        elif "signature_img" in udct:
            # Simple image signature, stretches to fit the box
            # image to render is contained in udct['signature_image']
            annotation.add_image(udct["signature_img"], "Image")
            annotation.set_signature_appearance(
                [
                    "image",
                    "Image",
                    0,
                    0,
                    x2 - x1,
                    y2 - y1,
                    udct.get("signature_img_distort", True),
                    udct.get("signature_img_centred", False),
                ]
            )
        elif "signature_appearance" in udct:
            # Adobe-inspired signature with text and images
            # Parameters are contained in udct['signature_appearance']
            # If a field is included in the display list, that field
            # will be contained in the annotation.
            #
            # Text and border are the colour specified by outline,
            # and border is the the inset distance from the outer
            # edge of the annotation.  The R G B values range between
            # 0 and 1.
            #
            # Icon is an image to display above the background and
            # border at the left-hand side of the anntoation.  If
            # there is no text, it is centred.
            #
            # The text block is left-aligned to the right of the icon
            # image.  If there is no image, the text is left-aliged
            # with the left-hand border of the annotation
            #
            # display fields:
            #   CN, DN, date, contact, reason, location
            #
            # Dict format:
            #   appearance = dict(
            #       background = Image with alpha / None,
            #       icon = Image with alpha / None,
            #       labels = bool,
            #       display = list,
            #       software = str,
            #       outline = [R, G, B],
            #       border = int,
            #       )
            sig = {}
            for f in ("background", "icon", "labels", "border", "outline"):
                if f in udct["signature_appearance"]:
                    sig[f] = udct["signature_appearance"][f]

            toggles = udct["signature_appearance"].get("display", [])
            for f in ("contact", "reason", "location", "contact", "signingdate"):
                if f in toggles:
                    sig[f] = udct.get(f, "{} unknown".format(f))
            if "date" in toggles:
                sig["date"] = udct["signingdate"]
            if "CN" in toggles:
                from cryptography.x509 import ObjectIdentifier

                sig["CN"] = cert.subject.get_attributes_for_oid(
                    ObjectIdentifier("2.5.4.3")
                )[0].value
            if "DN" in toggles:
                sig["DN"] = cert.subject.rfc4514_string()
            annotation.simple_signature(sig)
        else:
            # Manual signature annotation creation
            #
            # Make your own appearance with an arbitrary number of
            # images and fonts
            if "manual_images" in udct:
                for name, img in udct["manual_images"].items():
                    annotation.add_image(img, name=name)
            if "manual_fonts" in udct:
                for name, path in udct["manual_fonts"].items():
                    annotation.add_ttf_font(path, name=name)
            annotation.add_default_font()
            annotation.set_signature_appearance(*udct["signature_manual"])

        pdfa = annotation.as_pdf_object(identity(), page=page0ref)
        objapn = self._extend(pdfa["/AP"]["/N"])
        objapnref = self._addObject(objapn)

        objap = po.DictionaryObject()
        objap[po.NameObject("/N")] = objapnref
        obj13.update(
            {
                po.NameObject("/Rect"): po.ArrayObject(
                    [
                        po.FloatObject(x1),
                        po.FloatObject(y1),
                        po.FloatObject(x2),
                        po.FloatObject(y2),
                    ]
                ),
                po.NameObject("/AP"): objap,
                # po.NameObject("/SM"): po.createStringObject("TabletPOSinline"),
            }
        )

        page0 = page0ref.getObject()
        if new_13:
            annots = po.ArrayObject([obj13ref])
            if "/Annots" in page0:
                page0annots = page0["/Annots"]
                if isinstance(page0annots, po.IndirectObject):
                    annots.insert(0, page0annots)
                elif isinstance(page0annots, po.ArrayObject):
                    annots = page0annots
                    annots.append(obj13ref)
        else:
            annots = page0["/Annots"]
        page0.update({po.NameObject("/Annots"): annots})
        self._objects[page0ref.idnum - 1] = page0

    def makepdf(self, prev, udct, algomd, zeros, cert, **params):
        catalog = prev.trailer["/Root"]
        size = prev.trailer["/Size"]
        pages = catalog["/Pages"].getObject()
        page0ref = prev.getPage(udct.get("sigpage", 0)).indirectRef

        self._objects = []
        while len(self._objects) < size - 1:
            self._objects.append(None)

        if params['mode'] == 'timestamp':
            # deal with extensions
            if "/Extensions" not in catalog:
                extensions = po.DictionaryObject()
            else:
                extensions = catalog["/Extensions"]

            if "/ESIC" not in extensions:
                extensions.update(
                    {
                        po.NameObject("/ESIC"): po.DictionaryObject(
                            {
                                po.NameObject("/BaseVersion"): po.NameObject("/1.7"),
                                po.NameObject("/ExtensionLevel"): po.NumberObject(1),
                            }
                        )
                    }
                )
                catalog.update({po.NameObject("/Extensions"): extensions})
            else:
                esic = extensions["/ESIC"]
                major, minor = esic["/BaseVersion"].lstrip("/").split(".")
                if int(major) < 1 or int(minor) < 7:
                    esic.update(
                        {
                            po.NameObject("/BaseVersion"): po.NameObject("/1.7"),
                            po.NameObject("/ExtensionLevel"): po.NumberObject(1),
                        }
                    )

            SubFilter=po.NameObject("/ETSI.CAdES.detached")
        else:
            SubFilter=po.NameObject("/adbe.pkcs7.detached")

        # obj12 is the digital signature
        obj12, obj12ref = self._make_signature(
            Type=po.NameObject("/Sig"),
            SubFilter=SubFilter,
            Contents=UnencryptedBytes(zeros),
        )

        obj12.update(
            {
                po.NameObject("/Prop_Build"): pdf.DictionaryObject(
                    {
                        po.NameObject("/App"): pdf.DictionaryObject(
                            {
                                po.NameObject("/Name"): po.NameObject(
                                    "/" + udct.get("application", "endesive")
                                )
                            }
                        ),
                    }
                ),
            }
        )

        if params["mode"] == "timestamp":
            # obj12 is a timestamp this time
            obj12.update(
                {
                    po.NameObject("/Type"): po.NameObject("/DocTimeStamp"),
                    po.NameObject("/SubFilter"): po.NameObject("/ETSI.RFC3161"),
                    po.NameObject("/V"): po.NumberObject(0),
                }
            )
        d12 = {}
        for k, v in (
            ('/Name', 'name'),
            ('/Contact', 'contact'),
            ('/Location', 'location'),
            ('/Reason', 'reason'),
        ):
            v = udct.get(v, None)
            if v is not None:
                d12[po.NameObject(k)] = po.createStringObject(v)
        if 1 or params.get("use_signingdate"):
            d12[po.NameObject("/M")] = po.createStringObject(udct["signingdate"])
        if d12:
            obj12.update(d12)

        # obj13 is a combined AcroForm Sig field with Widget annotation
        new_13 = True
        # obj13 = po.DictionaryObject()
        if udct.get("signform", False):
            # Attaching signature to existing field in AcroForm
            if "/AcroForm" in catalog:
                form = catalog["/AcroForm"].getObject()
                if "/Fields" in form:
                    fields = form["/Fields"].getObject()
                    obj13ref = [
                        f
                        for f in fields
                        if f.getObject()["/T"] == udct.get("sigfield", "Signature1")
                    ][0]
                    obj13 = obj13ref.getObject()
                    self._objects[obj13ref.idnum - 1] = obj13
                    new_13 = False

        # box is coordinates of the annotation to fill
        box = udct.get("signaturebox", None)

        if new_13:
            obj13, obj13ref = self._make_sig_annotation(
                F=po.NumberObject(udct.get("sigflagsft", 132)),
                T=EncodedString(udct.get("sigfield", "Signature1")),
                Vref=obj12ref,
                Pref=page0ref,
            )
        else:
            # original obj13 is a merged SigField/SigAnnot
            # Setting /V on the AcroForm field sets the signature
            # for the field
            obj13.update({po.NameObject("/V"): obj12ref})
            # fill the existing signature field annotation,
            # ignore any other location
            if "/Rect" in obj13:
                box = [float(f) for f in obj13["/Rect"]]

        # add an annotation if there is a field to fill
        if box is not None:
            self.addAnnotation(cert, udct, box, page0ref, obj13, obj13ref, new_13)
        else:
            # invisible signature
            objap = po.DictionaryObject()
            objapref = self._addObject(objap)
            objform = po.DictionaryObject({
                po.NameObject("/Length"): po.NumberObject(0),
                po.NameObject("/Type"): po.NameObject("/XObject"),
                po.NameObject("/Subtype"): po.NameObject("/Form"),
                po.NameObject("/Bbox"): po.ArrayObject([
                    po.FloatObject(0.0),
                    po.FloatObject(0.0),
                    po.FloatObject(0.0),
                    po.FloatObject(0.0),
                ]),
            })
            objformref = self._addObject(objform)
            objap.update({po.NameObject("/N"): objformref})
            obj13.update({po.NameObject("/AP"): objapref})

        if udct.get("sigandcertify", False) and "/Perms" not in catalog:
            obj10 = po.DictionaryObject()
            obj10ref = self._addObject(obj10)
            obj11 = po.DictionaryObject()
            obj11ref = self._addObject(obj11)
            obj14 = po.DictionaryObject()
            obj14ref = self._addObject(obj14)
            obj14.update({po.NameObject("/DocMDP"): obj12ref})
            obj10.update(
                {
                    po.NameObject("/Type"): po.NameObject("/TransformParams"),
                    po.NameObject("/P"): po.NumberObject(udct.get("sigflags", 3)),
                    po.NameObject("/V"): po.NameObject("/1.2"),
                }
            )
            obj11.update(
                {
                    po.NameObject("/Type"): po.NameObject("/SigRef"),
                    po.NameObject("/TransformMethod"): po.NameObject("/DocMDP"),
                    po.NameObject("/DigestMethod"): po.NameObject("/" + algomd.upper()),
                    po.NameObject("/TransformParams"): obj10ref,
                }
            )
            obj12[po.NameObject("/Reference")] = po.ArrayObject([obj11ref])
            catalog[po.NameObject("/Perms")] = obj14ref

        if "/AcroForm" in catalog:
            form = catalog["/AcroForm"].getObject()
            if "/Fields" in form:
                fields = form["/Fields"]
                old_field_names = [f.getObject()["/T"] for f in fields]
            else:
                fields = po.ArrayObject()
                old_field_names = []
            if udct.get("auto_sigfield", False) and obj13["/T"] in old_field_names:
                name_base = udct.get("sigfield", "Signature1")
                checklist = [
                    f[len(name_base) :]
                    for f in old_field_names
                    if f.startswith(name_base)
                ]
                for i in range(1, len(checklist) + 1):
                    suffix = "_{}".format(i)
                    if suffix in checklist:
                        continue

                    new_name = "{}{}".format(name_base, suffix)
                    obj13.update({po.NameObject("/T"): EncodedString(new_name)})
                    break

            old_flags = int(form.get("/SigFlags", 0))
            new_flags = int(form.get("/SigFlags", 0)) | udct.get("sigflags", 3)
            if new_13:
                fields.append(obj13ref)
                form.update(
                    {
                        po.NameObject("/Fields"): fields,
                        po.NameObject("/SigFlags"): po.NumberObject(new_flags),
                    }
                )
            elif new_flags > old_flags:
                form.update({po.NameObject("/SigFlags"): po.NumberObject(new_flags)})
            formref = catalog.raw_get("/AcroForm")
            if isinstance(formref, po.IndirectObject):
                self._objects[formref.idnum - 1] = form
                form = formref
        else:
            form = po.DictionaryObject()
            form.update(
                {
                    po.NameObject("/Fields"): po.ArrayObject([obj13ref]),
                    po.NameObject("/SigFlags"): po.NumberObject(
                        udct.get("sigflags", 3)
                    ),
                }
            )
        catalog[po.NameObject("/AcroForm")] = form

        if "/Metadata" in catalog:
            catalog[po.NameObject("/Metadata")] = catalog.raw_get("/Metadata")

        x_root = prev.trailer.raw_get("/Root")
        self._objects[x_root.idnum - 1] = catalog
        self.x_root = po.IndirectObject(x_root.idnum, 0, self)
        self.x_info = prev.trailer.get("/Info")

    def sign(
        self,
        datau,
        udct,
        key,
        cert,
        othercerts,
        algomd,
        hsm,
        timestampurl=None,
        timestampcredentials=None,
        timestamp_req_options=None,
        mode="sign",
        ocspurl=None,
        ocspissuer=None,
    ):
        startdata = len(datau)

        fi = io.BytesIO(datau)

        # read end decrypt
        prev = pdf.PdfFileReader(fi)
        if prev.isEncrypted:
            rc = prev.decrypt(udct["password"])
        else:
            rc = 0

        # digest method must remain unchanged from prevoius signatures
        obj = prev.trailer
        for k in ("/Root", "/Perms", "/DocMDP", "/Reference"):
            if k in obj:
                obj = obj[k]
                if isinstance(obj, po.ArrayObject):
                    obj = obj[0]
                obj = obj.getObject()
            else:
                obj = None
                break
        if obj is not None:
            algomd = obj["/DigestMethod"][1:].lower()

        # produce smaller signatures, but must be signed twice
        aligned = udct.get("aligned", 4096 if isinstance(key, ec.EllipticCurvePrivateKey) else 0)
        if aligned:
            zeros = b"00" * aligned
        else:
            md = getattr(hashlib, algomd)().digest()
            if mode == "timestamp":
                contents = signer.timestamp(
                    None,
                    algomd,
                    timestampurl,
                    timestampcredentials,
                    timestamp_req_options,
                    prehashed=md,
                )[0]["values"][0].dump()
            else:
                attrs = udct.get("attrs", True)
                pss = udct.get("pss", False)
                contents = signer.sign(
                    None,
                    key,
                    cert,
                    othercerts,
                    algomd,
                    attrs,
                    md,
                    hsm,
                    pss,
                    timestampurl,
                    timestampcredentials,
                    timestamp_req_options,
                    ocspurl,
                    ocspissuer,
                )
            zeros = contents.hex().encode("utf-8")

        params = {"mode": mode}
        if not timestampurl:
            params["use_signingdate"] = True

        self.makepdf(prev, udct, algomd, zeros, cert, **params)

        # if document was encrypted, encrypt this version too
        if prev.isEncrypted:
            self.encrypt(prev, udct["password"], rc)
        else:
            self._encrypt_key = None

        # ID[0] is used in password protection, must be unchanged
        ID = prev.trailer.get("/ID", None)
        if ID is None:
            ID = udct.get("id") or hashlib.md5(repr(time.time()).encode()).digest()
        else:
            ID = ID.getObject()[0].original_bytes
        newID = udct.get("newid", repr(random.random()))
        self._ID = po.ArrayObject(
            [
                po.ByteStringObject(ID),
                po.ByteStringObject(hashlib.md5(newID.encode()).digest()),
            ]
        )

        fo = io.BytesIO()
        self.write(fo, prev, startdata)
        datas = fo.getvalue()

        br = [0, 0, 0, 0]
        bfrom = (b"[ " + b" ".join([WNumberObject.Format] * 4) + b" ]") % tuple(br)

        pdfbr1 = datas.find(zeros)
        pdfbr2 = pdfbr1 + len(zeros)
        br = [
            0,
            startdata + pdfbr1 - 1,
            startdata + pdfbr2 + 1,
            len(datas) - pdfbr2 - 1,
        ]
        bto = b"[%d %d %d %d]" % tuple(br)
        bto += b" " * (len(bfrom) - len(bto))
        assert len(bfrom) == len(bto)
        datas = datas.replace(bfrom, bto, 1)

        md = getattr(hashlib, algomd)()
        md.update(datau)
        b1 = datas[: br[1] - startdata]
        b2 = datas[br[2] - startdata :]
        md.update(b1)
        md.update(b2)
        md = md.digest()

        if mode == "timestamp":
            contents = signer.timestamp(
                None,
                algomd,
                timestampurl,
                timestampcredentials,
                timestamp_req_options,
                prehashed=md,
            )[0]["values"][0].dump()
        else:
            attrs = udct.get("attrs", True)
            pss = udct.get("pss", False)
            contents = signer.sign(
                None,
                key,
                cert,
                othercerts,
                algomd,
                attrs,
                md,
                hsm,
                pss,
                timestampurl,
                timestampcredentials,
                timestamp_req_options,
                ocspurl,
                ocspissuer,
            )
        contents = contents.hex().encode("utf-8")
        if aligned:
            nb = len(zeros) - len(contents)
            contents += b"0" * nb
        assert len(zeros) == len(contents)

        datas = datas.replace(zeros, contents, 1)

        return datas


def timestamp(
    datau,
    udct,
    algomd="sha1",
    timestampurl=None,
    timestampcredentials=None,
    timestamp_req_options=None,
):
    """
    parameters:
        datau: pdf bytes being timestamped
        udct: dictionary with signing parameters
            Same paramaters as the sign method.  The following parameters
            should not be used as they are either unused or conflict with
            normal timestamping:

            - signform
            - sigandcertify
            - contact
            - location
            - signingdate
            - reason

            Timestamping also does not typically use signature appearances.
            The following appearance options should probably be omitted,
            but do not conflict with the normal timestamping operations.

            - signaturebox
            - signature
            - signature_img
            - signature_img_distort
            - signature_img_centred
            - signature_appearance
            - signature_manual
            - manual_fonts
            - manual_images
        algomd:string                   default: sha1 - name of the hashing algorithm used to calculate
                                            the hash of the document being signed e.g.: sha1, sha256, sha384, sha512, ripemd160
        hsm: an instance of endesive.hsm.HSM class used to sign using a hardware token or None
        timestampurl: timestamp server URL or None
        timestampcredentials:Dict username and password for authentication against timestamp server. Default: None
        timestamp_req_options: Dict to set options to the POST http call against the timestamp server. Default: None

    returns: bytes ready for writing after unsigned pdf document containing its electronic timestamp
    """

    cls = SignedData()
    return cls.sign(
        datau,
        udct,
        None,  # key,
        None,  # cert,
        None,  # othercerts,
        algomd,
        None,  # hsm,
        timestampurl,
        timestampcredentials,
        timestamp_req_options,
        mode="timestamp",
    )


def sign(
    datau,
    udct,
    key,
    cert,
    othercerts,
    algomd="sha1",
    hsm=None,
    timestampurl=None,
    timestampcredentials=None,
    timestamp_req_options=None,
    ocspurl=None,
    ocspissuer=None,
):
    """
    parameters:
        datau: pdf bytes being signed
        udct: dictionary with signing parameters
            aligned: int                if 0 then precompute size of signature, but first fake data will be signed
                                        !=0 number of hexbytes (00) reserved for signature,
                                            must be equal or greather than hex representation of signature
                                            probably 16384 will be sufficient ....
            sigflags: int               default:3 1,2,3 - flags for acroform
            sigflagsft: int             default:132 - flags for annotation widget from pdf 12.5.3
            sigpage: int                default:0 - page on which signature should appear
            sigfield: string            default:Signature1
            auto_sigfield: bool         default:False
                                                False - do not check for sigfield name conflicts
                                                True  - append and increment suffix to sigfield when a field
                                                        by the name of sigfield already exists in AcroForm
            signform: bool              default:False
                                                False - do not fill in an existing form's signature field
                                                True  - attach signature to the pre-existing Sig field by the
                                                        the name provided by sigfield in AcroForm.  Use with
                                                        signature or signature_img, will fill the signature
                                                        field's widget, ignoring signaturebox.
            sigandcertify: bool         default:False
                                                False - sign only document
                                                True  - sign and certify document
            signaturebox: box|None      default:None - where to put signature image/string on selected page
            signature: string                   if box is not None then it should be latin1 encodable string
            signature_img: string|pil_image     if box is not None and string is None then it should be
                                                    pil image instance or
                                                    image file name or
                                                    byte array of image
            signature_img_distort: bool default:True
                                                True  - do not maintain image aspect ratio, fill the
                                                        entire bounding box, distorting the image.
                                                        This is set to True to match the behaviour
                                                        of previous versions of endesive
                                                False - maintain image aspect ratio when fitting
                                                        image into the bounding box
            signature_img_centred: bool default:True
                                                True  - centre images within the bounding box when
                                                        image aspect ratio does not match that of
                                                        the bounding box
                                                False - do not centre image, position it at the
                                                        lower-left of the bounding box
            signature_appearance: dict  if box is not None then render a signature appearance that is
                                            configured by the dict.  See below for configuration
            signature_manual: list      if box is not None then render a manually-created signature
                                        apperance using the directives contained in this list
            manual_fonts: dict          fonts required by the manual signature appearance
            manual_images: dict         images required by the manual signature appearance
            contact: string             required info about the person signing the document
            location:string             required info about location of the person signing the document
            signingdate: string         required info about signing time eg: now.strftime('D:%Y%m%d%H%M%S+00\'00\'')
            reason: string              required info about reason for signing the document
            password: string            required if the document is password protected, signing it also requires that password
            text: dict                  text attributes
                                            wraptext=True, fontsize:12, textalign:'left', linespacing:1.2
            application: string         optional application name in advanced signature properties dialog in Acrobat Reader
            pss: boolen                 optional choosen signature scheme RSASSA-PSS (true) vs. PKCS1-v1_5 (false/default)
        key: cryptography.hazmat.backends.openssl.rsa._RSAPrivateKey - private key used to sign the document
        cert: cryptography.x509.Certificate - certificate associated with the key
        othercerts: list of cryptography.x509.Certificate to be saved with the signed document,
            e.g.: a list of intermediate certificates used to confirm the authenticity of the certificate used in the signature
        algomd:string                   default: sha1 - name of the hashing algorithm used to calculate
                                            the hash of the document being signed e.g.: sha1, sha256, sha384, sha512, ripemd160
        hsm: an instance of endesive.hsm.HSM class used to sign using a hardware token or None
        timestampurl: timestamp server URL or None
        timestampcredentials:Dict username and password for authentication against timestamp server. Default: None
        timestamp_req_options: Dict to set options to the POST http call against the timestamp server. Default: None
        ocsppurl: ocsp server URL or None
        ocspissuer: certificate of issuer or None

    returns: bytes ready for writing after unsigned pdf document containing its electronic signature

    Signature appearance:

    Adobe-inspired signature with text and images Parameters are contained
    in udct['signature_appearance']

    If a field is included in the display list, that field will be included
    in the annotation.  Its value is based upon what is in udct and the cert.
    Text and border are the colour specified by outline, and border is the
    the inset distance from the outer edge of the annotation.  The R G B
    values range between 0 and 1.

    Icon is an image to display above the background and border at the
    left-hand side of the anntoation.  If there is no text, it is centred.

    The text block is left-aligned to the right of the icon image.  If there
    is no image, the text is left-aliged with the left-hand border of the
    annotation

    display fields:
        CN, DN, date, contact, reason, location

    Dict format:
      appearance = dict(
          background = Image with alpha / None,
          icon = Image with alpha / None,
          labels = bool,
          display = list,
          software = str,
          outline = [R, G, B],
          border = int,
          )

    Signature manual:

    Manually set up your own signature appearance with arbitrary fonts and
    images.  The full list of directives can be seen in the template dict
    in endesive.pdf.PyPDF2_annotate.annotations.signatureaSignatureAppearance

    Coordinates are based off of the lower-left corner of the annotation

    manual_images is a dict of images that will be included in the annotation
    The keys are strings, and are the names to use when referring to the them
    in 'image' directives.

    manual_fonts is a dict of TrueType fonts that will be included in the
    annotation.  The keys are strings, and are the names to use when referring
    to the them in 'image' directives.  The values are the paths to the font
    files.  Do not use the name 'default' in this dict; it is treated
    specially.

    """
    cls = SignedData()
    return cls.sign(
        datau,
        udct,
        key,
        cert,
        othercerts,
        algomd,
        hsm,
        timestampurl,
        timestampcredentials,
        timestamp_req_options,
        "sign",
        ocspurl,
        ocspissuer,
    )



# -*- coding: utf-8 -*-
"""
    Text Utils
    ~~~~~~~~~~

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""


def unshift_token(text):
    """Remove a token from the front of a string.

    :param str text:
    :returns: {'text': str, 'separator': str, 'remainder': str}
    """
    if len(text) == 0:
        return {'text': text, 'separator': '', 'remainder': ''}

    token = ''
    for i in range(0, len(text)):
        char = text[i]
        if (char == ' ' and (len(token) >= 1 and token[i - 1] == ' ')):
            token += char
        elif (char == ' ' and len(token) == 0):
            token += char
        elif char == ' ':
            return {'text': token, 'separator': ' ', 'remainder': text[i + 1:]}
        elif char == '\n':
            return {
                'text': token,
                'separator': '\n',
                'remainder': text[i + 1:],
            }
        elif (len(token) >= 1 and token[i - 1] == ' '):
            return {
                'text': token,
                'separator': '',
                'remainder': text[len(token):],
            }
        else:
            token += char

    return {'text': token, 'separator': '', 'remainder': ''}


def unshift_line(text, measure, max_length):
    """Remove a line of text from a string.

    :param str text: text to be broken
    :param func measure: function that takes a string and returns its width
    :param int max_length: max width of each line
    :returns: {'text': str, 'remainder': str}
    """
    line = ''
    token = {'text': '', 'separator': '', 'remainder': text}
    while True:
        token = unshift_token(token['remainder'])
        token_text = token['text']
        remainder = token['remainder']
        separator = token['separator']
        if len(line) == 0:
            if len(token_text) > 0:
                # This allows us to add partial tokens for the first token
                for char in token_text:
                    if measure(line + char) > max_length:
                        line = char if len(line) == 0 else line
                        return {
                            'text': line,
                            'remainder': text[len(line):],
                        }
                    else:
                        line += char
                if separator == '\n':
                    return {'text': line, 'remainder': remainder}

                line += separator
            else:
                return {
                    'text': line,
                    'remainder': text[len(line) + len(separator):],
                }
        else:
            if measure(line + token_text) <= max_length:
                line += token_text
                if separator == '\n' or remainder == '':
                    return {'text': line, 'remainder': remainder}
                else:
                    line += separator
            else:
                return {'text': line, 'remainder': text[len(line):]}


def get_wrapped_lines(text, measure, max_length):
    """Break a string of text into lines wrapped to max_length.

    The algorithm is the same one used in the PGBS TextElement in web-viewer,
    to maintain consistency in line breaks.

    :param str text: text to be broken
    :param func measure: function that takes a string and returns its width
    :param int max_length: max width of each line
    :returns: list of strings
    """
    line = unshift_line(text, measure, max_length)
    lines = [line['text']]
    while (len(line['remainder']) > 0):
        line = unshift_line(line['remainder'], measure, max_length)
        lines.append(line['text'])
    return lines



# -*- coding: utf-8 -*-
"""
    Rectangular Annotations
    ~~~~~~~~~~~~~~~~~~~~~~~
    Annotations defined by a width and a height: Square, Circle

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
from .base import Annotation
from .base import make_border_dict
from ..pdfrw import PdfArray, PdfName
from ..config.appearance import set_appearance_state
from ..config.appearance import stroke_or_fill
from ..graphics import Bezier
from ..graphics import Close
from ..graphics import ContentStream
from ..graphics import Line
from ..graphics import Move
from ..graphics import quadratic_to_cubic_bezier
from ..graphics import Rect
from ..graphics import Restore
from ..graphics import Save


class RectAnnotation(Annotation):
    """Abstract annotation that defines its location on the document with a
    width and a height.
    """

    def make_rect(self):
        stroke_width = self._appearance.stroke_width
        L = self._location
        return [
            L.x1 - stroke_width,
            L.y1 - stroke_width,
            L.x2 + stroke_width,
            L.y2 + stroke_width,
        ]

    def add_additional_pdf_object_data(self, obj):
        A = self._appearance
        obj[PdfName("BS")] = make_border_dict(A)
        obj[PdfName("C")] = A.stroke_color
        if A.fill:
            obj[PdfName("IC")] = A.fill
        padding = A.stroke_width / 2.0
        obj[PdfName("RD")] = PdfArray([padding, padding, padding, padding])


class Square(RectAnnotation):
    subtype = "Square"

    def make_appearance_stream(self):
        L = self._location
        A = self._appearance
        stream = ContentStream([Save()])

        set_appearance_state(stream, A)
        stream.add(Rect(L.x1, L.y1, L.x2 - L.x1, L.y2 - L.y1))
        stroke_or_fill(stream, A)
        stream.add(Restore())

        # TODO dash array
        return stream


def add_rounded_rectangle(stream, x, y, width, height, rx, ry):
    """Creates a rounded rectangle and adds it to the content stream.

    :param ContentStream stream:
    :param float x1:
    :param float y1:
    :param float width:
    :param float height:
    :param float rx: x radius of the rounded corners
    :param float ry: y radius of the rounded corners
    """
    stream.add(Move(x + rx, y))
    stream.add(Line(x + width - rx, y))
    stream.add(
        quadratic_to_cubic_bezier(
            start_x=(x + width - rx),
            start_y=y,
            control_x=(x + width),
            control_y=y,
            end_x=(x + width),
            end_y=(y + ry),
        )
    )
    stream.add(Line(x + width, y + height - ry))
    stream.add(
        quadratic_to_cubic_bezier(
            start_x=(x + width),
            start_y=(y + height - ry),
            control_x=(x + width),
            control_y=(y + height),
            end_x=(x + width - rx),
            end_y=(y + height),
        )
    )
    stream.add(Line(x + rx, y + height))
    stream.add(
        quadratic_to_cubic_bezier(
            start_x=(x + rx),
            start_y=(y + height),
            control_x=x,
            control_y=(y + height),
            end_x=x,
            end_y=(y + height - ry),
        )
    )
    stream.add(Line(x, y + ry))
    stream.add(
        quadratic_to_cubic_bezier(
            start_x=x,
            start_y=(y + ry),
            control_x=x,
            control_y=y,
            end_x=(x + rx),
            end_y=y,
        )
    )
    stream.add(Close())


def add_bezier_circle(stream, x1, y1, x2, y2):
    """Create a circle from four bezier curves and add it to the content stream,
    since PDF graphics is missing an ellipse primitive.

    :param ContentStream stream:
    :param float x1:
    :param float y1:
    :param float x2:
    :param float y2:
    """
    left_x = x1
    right_x = x2
    bottom_x = left_x + (right_x - left_x) / 2.0
    top_x = bottom_x

    bottom_y = y1
    top_y = y2
    left_y = bottom_y + (top_y - bottom_y) / 2.0
    right_y = left_y

    cp_offset = 0.552284749831
    # Move to the bottom of the circle, then four curves around.
    # https://stackoverflow.com/questions/1734745/how-to-create-circle-with-b%C3%A9zier-curves
    stream.add(Move(bottom_x, bottom_y))
    stream.add(
        Bezier(
            bottom_x + (right_x - bottom_x) * cp_offset,
            bottom_y,
            right_x,
            right_y - (right_y - bottom_y) * cp_offset,
            right_x,
            right_y,
        )
    )
    stream.add(
        Bezier(
            right_x,
            right_y + (top_y - right_y) * cp_offset,
            top_x + (right_x - top_x) * cp_offset,
            top_y,
            top_x,
            top_y,
        )
    )
    stream.add(
        Bezier(
            top_x - (top_x - left_x) * cp_offset,
            top_y,
            left_x,
            left_y + (top_y - left_y) * cp_offset,
            left_x,
            left_y,
        )
    )
    stream.add(
        Bezier(
            left_x,
            left_y - (left_y - bottom_y) * cp_offset,
            bottom_x - (bottom_x - left_x) * cp_offset,
            bottom_y,
            bottom_x,
            bottom_y,
        )
    )
    stream.add(Close())


class Circle(RectAnnotation):
    """Circles and Squares are basically the same PDF annotation but with
    different content streams.
    """

    subtype = "Circle"

    def make_appearance_stream(self):
        L = self._location
        A = self._appearance

        stream = ContentStream([Save()])
        set_appearance_state(stream, A)
        add_bezier_circle(stream, L.x1, L.y1, L.x2, L.y2)
        stroke_or_fill(stream, A)
        stream.add(Restore())

        return stream



# -*- coding: utf-8 -*-
"""
    True Type Fonts Utility Class
    ~~~~~~~~~~

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
from fontTools.ttLib import TTFont

from .font_metrics import FontMetrics


_FONT_CACHE = {}


def get_true_type_font(path, font_name, font_size=None):
    """Helper to avoid having to reload font from disk multiple times in a session

    :param str path: path to .ttf font file
    :param str font_name: name of the font to be included in the PDF file
    :param int|None font_size:
    :returns TrueTypeFont:
    """
    key = (path, font_name, font_size)
    font = _FONT_CACHE.get(key)
    if font is not None:
        return font

    font = TrueTypeFont(path, font_name, font_size)
    _FONT_CACHE[key] = font
    return font


class TrueTypeFont:
    """
    Used to load a true type font and calculate font metrics from it that are
    needed to embed the font program in a PDF.
    """

    def __init__(self, path, font_name, font_size=None):
        self.ttfPath = path
        self._ttfFont = TTFont(self.ttfPath)
        # Subsetted fonts have 6 random letters prepended to their names
        # See section 9.6.4 - Font Subsets of the PDF 1.7 Spec
        self.fontName = 'RXMLFT+' + font_name

        self.metrics = self._calculate(self._ttfFont)
        self._glyph_set = self._ttfFont.getGlyphSet()
        self._font_size = font_size

    def get_glyph_id(self, glyph_name):
        """
        Wrapper for getting a glyph name via font tools.
        :param glyph_name: The name of the glyph we're retrieving.
        :return: The corresponding glyph ID.
        """
        return self._ttfFont['glyf'].getGlyphID(glyph_name)

    def measure_text(self, text, font_size=None):
        """Measure a block of text using the font's metrics. If the text
        contains characters the font does not define, the .notdef character's
        width is used.

        :param str text: The text to measure
        :param int|None font_size: Font size (in em units) to scale
            measurements. If missing, self._font_size is used.
        :returns int: width of text
        """
        font_size = font_size if font_size is not None else self._font_size
        if font_size is None:
            raise ValueError('Font size must be specified')

        total_width = 0
        notdef_width = self._glyph_set['.notdef'].width
        for character in text:
            # If the cmap doesn't contain the character, this'll just return
            # None for the glyph and use .notdef
            glyph = self._glyph_set.get(self.metrics.cmap.get(ord(character)))
            total_width += glyph.width if glyph is not None else notdef_width
        # Scale total width by font size
        return total_width * font_size / self.metrics.unitsPerEm

    @staticmethod
    def _calculate(font):
        """
        Calculates metrics about our true type font.  These calculations are taken from
        previous work done in the mpdf/fpdf projects.
        :param font: The fonttools font object.
        :return: A FontMetrics object containing the calculated metrics.
        """
        # Font Header Table
        units_per_em = font['head'].unitsPerEm
        scale = 1000 / float(units_per_em)
        x_min = font['head'].xMin
        y_min = font['head'].yMin
        x_max = font['head'].xMax
        y_max = font['head'].yMax
        bbox = [
            (x_min * scale),
            (y_min * scale),
            (x_max * scale),
            (y_max * scale)
        ]

        # hhead metrics table, these seem to be used instead of OS/2 for compatibility reasons
        if 'hhea' in font:
            ascent = font['hhea'].ascent * scale
            descent = font['hhea'].descent * scale

        # OS/2 and Windows metrics table
        if 'OS/2' in font:
            us_weight_class = font['OS/2'].usWeightClass
            if not ascent:
                ascent = font['OS/2'].sTypoAscender * scale
            if not descent:
                descent = font['OS/2'].sTypoDescender * scale
            if font['OS/2'].version > 1:
                cap_height = font['OS/2'].sCapHeight * scale
            else:
                cap_height = ascent
        else:
            us_weight_class = 500
            if not ascent:
                ascent = y_max * scale
            if not descent:
                descent = y_min * scale
            cap_height = ascent
        stem_v = 50 + int(pow((us_weight_class / 65.0), 2))

        # Post table
        is_fixed_pitch = font['post'].isFixedPitch
        italic_angle = font['post'].italicAngle

        # hmtx - contains advance width and left side bearing for each glyph
        default_width = font['hmtx'].metrics['.notdef'][0]

        # Character map
        cmap = font['cmap'].getBestCmap()

        # Widths
        glyph_set = font.getGlyphSet()
        cids = [cid for cid in cmap]
        if len(cids) <= 0:
            raise MetricsParsingError("Couldn't find any characters in font")
        widths = TrueTypeFont._format_widths(glyph_set, cmap, cids)

        return FontMetrics(
            italicAngle=italic_angle,
            usWeightClass=us_weight_class,
            isFixedPitch=is_fixed_pitch,
            unitsPerEm=units_per_em,
            scale=scale,
            bbox=bbox,
            ascent=ascent,
            descent=descent,
            capHeight=cap_height,
            stemV=stem_v,
            defaultWidth=default_width,
            widths=widths,
            cmap=cmap,
        )

    @staticmethod
    def _format_widths(glyph_set, cmap, cids):
        """
        See Section 9.7.4.3 Glyph Metrics in CIDFonts
        This function will take a uniform list of widths and format it to the PDF compacted format.
        The widths use one of two formats for each segment.
        For varying widths: c [w1 w2 .. wn]
        For constant widths: cfirst clast w
        ie. [ 120 [ 400 325 500 ] 7080 8032 1000 ]
        Would have 120 be 400, 121 be 325, 122 be 500 and 7080 to 8032 all be 1000 width
        :return: A list of lists conforming to the PDF compacted format for widths.
        """
        cids_length = len(cids)
        if cids_length == 0:
            return []
        cids.sort()
        start = 0
        i = 1
        # See Section 9.7.4.3 Glyph Metrics in CIDFonts
        # The widths use one of two formats for each segment.
        # For varying widths: c [w1 w2 .. wn]
        # For constant widths: cfirst clast w
        # ie. [ 120 [ 400 325 500 ] 7080 8032 1000 ]
        # Would have 120 be 400, 121 be 325, 122 be 500 and 7080 to 8032 all be 1000 width
        widths = []
        while True:
            if i >= cids_length or (cids[i] - cids[i - 1] > 1):
                indices = [x for x in range(cids[start], cids[i - 1] + 1)]
                w = [glyph_set[cmap[index]].width for index in indices]

                if len(set(w)) == 1:
                    # Append a 'cfirst clast w' style width to our width segments
                    widths.append(cids[start])
                    widths.append(cids[i - 1])
                    widths.append(w[0])
                else:
                    # Append a 'c [w1 w2 .. wn]' style width to our width segments
                    widths.append(cids[start])
                    widths.append(w)

                # No more ranges
                if i >= cids_length:
                    break

                start = i
            i = i + 1
        return widths


class MetricsParsingError(Exception):
    pass



#!/usr/bin/env vpython3
# coding: utf-8
import unittest
import os
import stat
import subprocess
import sysconfig
import datetime
import base64
import email

from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
from email.mime.base import MIMEBase

from asn1crypto import pem as asn1pem
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import pkcs12
from endesive import hsm, signer, verifier
import PyKCS11 as PK11

import test_cert

tests_root = os.path.dirname(__file__)
fixtures_dir = os.path.join(tests_root, 'fixtures')

def fixture(fname):
    return os.path.join(fixtures_dir, fname)

dllpath = os.path.join(sysconfig.get_config_var('LIBDIR'), 'softhsm/libsofthsm2.so')

os.makedirs(os.path.join(fixtures_dir, 'softhsm2'), exist_ok=True)
os.environ['SOFTHSM2_CONF'] = fixture('softhsm2.conf')
open(fixture('softhsm2.conf'), 'wt').write('''\
log.level = DEBUG
directories.tokendir = %s/softhsm2/
objectstore.backend = file
slots.removable = false
''' % fixtures_dir)

class HSM(hsm.HSM):
    def main(self):
        cakeyID = bytes((0x1,))
        rec = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, cakeyID)])
        if len(rec) == 0:
            label = 'hsm CA'
            self.gen_privkey(label, cakeyID)
            self.ca_gen(label, cakeyID, 'hsm CA')

        keyID = bytes((0x66,0x66,0x90))
        rec = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, keyID)])
        if len(rec) == 0:
            label = 'hsm USER 1'
            self.gen_privkey(label, keyID)
            self.ca_sign(keyID, label, 0x666690, "hsm USER 1", 365, cakeyID)

        self.cert_export(fixture('cert-hsm-ca'), cakeyID)
        self.cert_export(fixture('cert-hsm-user1'), keyID)

def compose(From, To, Subject, Body, Attachment, signer):
    # create message object instance
    msg = MIMEMultipart(_subtype="signed", micalg="SHA1", protocol="application/pkcs7-signature")

    # setup the parameters of the message
    msg['From'] = From
    msg['To'] = To
    msg['Subject'] = Subject
    msg['Date'] = email.utils.format_datetime(datetime.datetime.now())
    msg.preamble = "This is a multipart message in MIME format."

    env = MIMEMultipart(_subtype='mixed')
    body = MIMEText(Body.decode())
    #del body['MIME-Version']
    env.attach(body)

    with open(Attachment, 'rb') as fp:
        app = MIMEApplication(fp.read(), _subtype="txt")
    app.add_header('content-disposition', 'attachment', filename=Attachment)
    env.attach(app)

    msg.attach(env)

    sig = MIMEBase(_maintype='application', _subtype='pkcs7-signature', name="smime.p7s")
    sig.add_header('Content-Disposition', 'attachment', filename='smime.p7s')
    sig.add_header('Content-Transfer-Encoding', 'base64')
    datau = env.as_string().encode()
    #open(fixture('test_ssh_sign-sign.txt'),'wb').write(datau)
    sig.set_payload(signer(datau))
    #del sig['MIME-Version']
    msg.attach(sig)

    return msg, env, sig


def sign(datau, key, cert, othercerts, hashalgo, hsm):
    datau = datau.replace(b'\n', b'\r\n')
    datas = signer.sign(datau, key, cert, othercerts, hashalgo, attrs=True, pss=False, hsm=hsm)
    return base64.encodebytes(datas)


class HSMTests(unittest.TestCase):
    def test_base(self):
        cls = hsm.BaseHSM()
        try:
            cls.certificate()
        except NotImplementedError:
            pass
        try:
            cls.sign(None, None, None)
        except NotImplementedError:
            pass

    def test_create(self):
        cls = HSM(dllpath)
        cls.create("endesieve", "secret1", "secret2")
        cls.login("endesieve", "secret1")
        try:
            cls.main()
        finally:
            cls.logout()

    def test_load(self):
        cls = HSM(dllpath)
        cls.login("endesieve", "secret1")
        cakeyID = bytes((0x1,))
        cls.cert_load(cakeyID)
        keyID = bytes((0x66,0x66,0x90))
        cls.cert_load(keyID)
        cls.logout()

    def test_ssh_sign(self):
        key, cert, othercerts = test_cert.CA().pk12_load(test_cert.cert1_p12, '1234')

        agent = hsm.SSHAgentHSM(cert)

        # lookup the ssh fingerprint for the certificates public key
        keyid, _ = agent.certificate()
        keyfile = None
        try:
            # is the public key known to the ssh-agent yet?
            agent.key(keyid)
        except ValueError:
            agent.close()

            # we have to add the key to the ssh-agent
            # remove the key password, dump in traditional openssl so ssh-agent can add the key
            keyfile = fixture('demo2_user1.key.nopass.pem')
            with open(keyfile, 'wb') as fp:
                # dump the key
                fp.write(key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption()
                ))
            os.chmod(keyfile, 0o600)
            # pub file so ssh-add -d can be used
            pubfile = fixture('demo2_user1.key.nopass.pem.pub')
            with open(pubfile, 'wb') as fp:
                # convert the public key of the certificate to ssh public key format
                fp.write(cert.public_key().public_bytes(
                    encoding=serialization.Encoding.OpenSSH,
                    format=serialization.PublicFormat.OpenSSH
                ))

            subprocess.call(["ssh-add", keyfile])

            # reconnect the agent so the key is visible to paramiko
            agent = hsm.SSHAgentHSM(cert)

        with open(fixture('smime-unsigned.txt'), 'rb') as fp:
            datau = fp.read()

        msg, env, sig = compose(
            From='root+from@localhost',
            To='root+to@localhost',
            Subject='this is the subject',
            Body=datau,
            Attachment=fixture('smime-unsigned.txt'),
            signer=lambda data: sign(data, None, cert, othercerts, 'sha256', agent)
        )
        datas = msg.as_bytes(unixfrom=True)
        with open(fixture('smime-signed-hsm-ssh.txt'), 'wb') as fp:
            fp.write(datas)

        # we added, so we remove the key from ssh-agent
        keyfile = fixture('demo2_user1.key.nopass.pem')
        if keyfile:
            subprocess.call(['ssh-add', '-d', keyfile])

        agent.close()

    def test_ssh_verify(self):
        with open(test_cert.cert1_cert, 'rb') as fp:
            cert = fp.read()
        with open(fixture('smime-signed-hsm-ssh.txt'), 'rt') as fp:
            datas = fp.read()

        msg = email.message_from_string(datas)
        sig = None
        plain = msg.get_payload(0, False)
        for part in msg.walk():
            ct = part.get_content_type()
            #if ct == 'application/x-pkcs7-signature':
            #    sig = part.get_payload(decode=True)
            #    break
            if ct == 'application/pkcs7-signature':
                sig = part.get_payload(decode=True)
                break
        #if sig is None:
        #    raise ValueError('not signed email')

        datau = plain.as_string().encode()
        #open(fixture('test_ssh_sign-plain.txt'),'wb').write(datau)

        datau = datau.replace(b'\n', b'\r\n')
        (hashok, signatureok, certok) = verifier.verify(sig, datau, [cert,])
        assert hashok and signatureok and certok



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import io
from endesive import email


def main():
    trusted_cert_pems = (open('ca/demo2_ca.crt.pem', 'rb').read(),)

    for fname in (
            'smime-ssl-signed-attr.txt',
            'smime-ssl-signed-noattr.txt',
            'smime-signed-attr.txt',
            'smime-signed-attr-custom.txt',
            'smime-signed-noattr.txt',
            'smime-ssl-pss-signed.txt',
            'smime-signed-pss-oscrypto.txt',
            'smime-signed-pss.txt',
    ):
        print('*' * 20, fname)
        try:
            datae = io.open(fname, 'rt', encoding='utf-8').read()
        except:
            print('no such file')
            continue
        (hashok, signatureok, certok) = email.verify(datae, trusted_cert_pems)
        print('signature ok?', signatureok)
        print('hash ok?', hashok)
        print('cert ok?', certok)


main()



'''
RSA-SHA-256: RSA signature condition using SHA-256.

This RSA condition uses RSA-PSS padding with SHA-256.
The salt length is set equal the digest length of 32 bytes.

The public exponent is fixed at 65537 and the public
modulus must be between 128 (1017 bits) and
512 bytes (4096 bits) long.

RSA-SHA-256 is assigned the type ID 3. It relies on
the SHA-256 and RSA-PSS feature suites which corresponds
to a feature bitmask of 0x11.

            parameters = sigalgo["parameters"]
            salgo = parameters["hash_algorithm"].native["algorithm"].upper()
            mgf = getattr(
                padding, parameters["mask_gen_algorithm"].native["algorithm"].upper()
            )(getattr(hashes, salgo)())
            salt_length = parameters["salt_length"].native
            try:
                public_key.verify(
                    signature,
                    signedData,
                    padding.PSS(mgf, salt_length),
                    getattr(hashes, salgo)(),
                )
                signatureok = True
            except:
                signatureok = False
'''
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, rsa

def encrypt(message, public_key):
    # Hash the message
    digest = hashes.Hash(hashes.SHA256())
    digest.update(message)
    hash_digest = digest.finalize()

    # Encrypt the hash with the public key
    encrypted = public_key.encrypt(
        hash_digest,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    return encrypted

def decrypt(encrypted, private_key):
    # Decrypt the encrypted hash
    decrypted = private_key.decrypt(
        encrypted,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    return decrypted

def main():
    private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=2048, backend=default_backend()
    )
    public_key = private_key.public_key()

    # Data to encrypt
    message = b"Hello, world!"

    if 0:
        encrypted = encrypt(message, public_key)
        decrypted = decrypt(encrypted, private_key)
        digest = hashes.Hash(hashes.SHA256())
        digest.update(message)
        hash_digest = digest.finalize()
        print('ok?', hash_digest == decrypted)


    salt_length = 32 # length(hash_digest)
    signature = private_key.sign(
        message,
        padding.PSS(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            salt_length=salt_length,
        ),
        algorithm=hashes.SHA256(),
    )
    try:
        public_key.verify(
            signature,
            message,
            padding.PSS(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                salt_length=salt_length
            ),
            algorithm=hashes.SHA256(),
        )
        ok = True
    except:
        ok = False
    print('ok=', ok)

main()



#!/usr/bin/env vpython3
# coding: utf-8
import unittest
import os
import sys
import io
from subprocess import PIPE, Popen
from datetime import datetime

from cryptography import x509
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12, load_pem_private_key
from endesive import email

import hashlib
from asn1crypto import cms, algos, core, pem

import test_cert

tests_root = os.path.dirname(__file__)
fixtures_dir = os.path.join(tests_root, 'fixtures')


def fixture(fname):
    return os.path.join(fixtures_dir, fname)


class EMAILTests(unittest.TestCase):
    def test_email_signed_attr(self):
        p12 = test_cert.CA().pk12_load(test_cert.cert1_p12, '1234')

        with open(fixture('smime-unsigned.txt'), 'rb') as fh:
            datau = fh.read()
        datas = email.sign(datau,
            p12[0], p12[1], p12[2],
            'sha256',
            attrs=True
        )
        fname = fixture('smime-signed-attr.txt')
        with open(fname, 'wb') as fh:
            fh.write(datas)

        with open(test_cert.cert1_cert, 'rb') as fp:
            cert = fp.read()
        (hashok, signatureok, certok) = email.verify(datas.decode('utf8'), [cert,])
        assert hashok and signatureok and certok

        cmd = [
            'openssl', 'smime', '-verify',
            '-CAfile', test_cert.ca_cert,
            '-in', fname, '-inform', 'SMIME',
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        assert stderr == b'Verification successful\n'
        assert datau.replace(b'\n', b'\r\n') == stdout

    def test_email_signed_attr_pss(self):
        p12 = test_cert.CA().pk12_load(test_cert.cert1_p12, '1234')

        with open(fixture('smime-unsigned.txt'), 'rb') as fh:
            datau = fh.read()
        datas = email.sign(datau,
            p12[0], p12[1], p12[2],
            'sha512',
            attrs=True,
            pss=True
        )
        fname = fixture('smime-signed-attr-pss.txt')
        with open(fname, 'wb') as fh:
            fh.write(datas)

        with open(test_cert.cert1_cert, 'rb') as fp:
            cert = fp.read()
        (hashok, signatureok, certok) = email.verify(datas.decode('utf8'), [cert,])
        assert hashok and signatureok and certok

        cmd = [
            'openssl', 'cms', '-verify',
            '-signer', test_cert.cert1_cert,
            '-keyopt', 'rsa_padding_mode:pss', '-md', 'sha512',
            '-CAfile', test_cert.ca_cert,
            '-in', fname
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        # OpenSSL <= 1.1.1 outputs 'Verification successful'
        # OpenSSL >= 3.0.0 outputs 'CMS Verification successful'
        assert stderr == b'Verification successful\n' or stderr == b'CMS Verification successful\n'
        assert datau.replace(b'\n', b'\r\n') == stdout

    def test_email_signed_attr_custom(self):
        p12 = test_cert.CA().pk12_load(test_cert.cert1_p12, '1234')
        with open(fixture('smime-unsigned.txt'), 'rb') as fh:
            datau = fh.read()

        datau1 = datau.replace(b'\n', b'\r\n')
        hashalgo = 'sha256'
        signed_value = getattr(hashlib, hashalgo)(datau1).digest()
        attrs = [
            cms.CMSAttribute({
                'type': cms.CMSAttributeType('content_type'),
                'values': ('data',),
            }),
            cms.CMSAttribute({
                'type': cms.CMSAttributeType('message_digest'),
                'values': (signed_value,),
            }),
        ]

        datas = email.sign(datau,
            p12[0], p12[1], p12[2],
            'sha256',
            attrs=attrs
        )
        fname = fixture('smime-signed-attr-custom.txt')
        with open(fname, 'wb') as fh:
            fh.write(datas)

        cmd = [
            'openssl', 'smime', '-verify',
            '-CAfile', test_cert.ca_cert,
            '-in', fname, '-inform', 'SMIME',
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        assert stderr == b'Verification successful\n'
        assert datau.replace(b'\n', b'\r\n') == stdout

    def test_email_signed_noattr(self):
        p12 = test_cert.CA().pk12_load(test_cert.cert1_p12, '1234')
        with open(fixture('smime-unsigned.txt'), 'rb') as fh:
            datau = fh.read()
        datas = email.sign(datau,
            p12[0], p12[1], p12[2],
            'sha256',
            attrs=False
        )
        fname = fixture('smime-signed-noattr.txt')
        with open(fname, 'wb') as fh:
            fh.write(datas)

        cmd = [
            'openssl', 'smime', '-verify',
            '-CAfile', test_cert.ca_cert,
            '-in', fname, '-inform', 'SMIME',
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        assert stderr == b'Verification successful\n'
        assert datau.replace(b'\n', b'\r\n') == stdout

    def test_email_crypt(self):
        certs = (
            test_cert.CA().cert_load(test_cert.cert1_cert),
        )
        with open(fixture('smime-unsigned.txt'), 'rb') as fh:
            datau = fh.read()
        datae = email.encrypt(datau, certs, 'aes256_ofb')
        fname = fixture('smime-encrypted.txt')
        with open(fname, 'wt') as fh:
            fh.write(datae)

        key = test_cert.CA().key_load(test_cert.cert1_key, '1234')
        with io.open(fname, 'rt', encoding='utf-8') as fh:
            datae = fh.read()
        datad = email.decrypt(datae, key)

        assert datau == datad

    def _test_email_ssl_decrypt(self, algo, mode, oaep):
        certs = (
            test_cert.CA().cert_load(test_cert.cert1_cert),
        )

        with open(fixture('smime-unsigned.txt'), 'rb') as fh:
            datau = fh.read()

        datae = email.encrypt(datau, certs, algo+'_'+mode, oaep)
        fname = fixture('smime-encrypted-{}-{}-{}.txt'.format(algo, mode, oaep))
        with open(fname, 'wt') as fh:
            fh.write(datae)

        if 0:
            key = test_cert.CA().key_load(test_cert.cert1_key, '1234')
            datau = email.decrypt(datae, key)

        if not oaep:
            cmd = [
                'openssl', 'smime', '-decrypt',
                '-recip', test_cert.cert1_cert,
                '-inkey', test_cert.cert1_key,
                '-passin', 'pass:1234',
                '-in', fname,
            ]
        else:
            cmd = [
                'openssl', 'cms', '-decrypt',
                '-recip', test_cert.cert1_cert,
                '-inkey', test_cert.cert1_key,
                '-passin', 'pass:1234',
                '-in', fname,
            ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        assert stderr == b''
        if stdout != datau:
            lastbyte = stdout[-1]
            stdout = stdout[:len(stdout)-lastbyte]
        assert stdout == datau

    def test_email_ssl_decrypt_aes128_cbc_False(self):
        self._test_email_ssl_decrypt('aes128', 'cbc', False)

    def test_email_ssl_decrypt_aes192_cbc_False(self):
        self._test_email_ssl_decrypt('aes192', 'cbc', False)

    def test_email_ssl_decrypt_aes256_cbc_False(self):
        self._test_email_ssl_decrypt('aes256', 'cbc', False)

    def test_email_ssl_decrypt_aes128_ofb_False(self):
        self._test_email_ssl_decrypt('aes128', 'ofb', False)

    def test_email_ssl_decrypt_aes192_ofb_False(self):
        self._test_email_ssl_decrypt('aes192', 'ofb', False)

    def test_email_ssl_decrypt_aes256_ofb_False(self):
        self._test_email_ssl_decrypt('aes256', 'ofb', False)

    def test_email_ssl_decrypt_aes128_cbc_True(self):
        self._test_email_ssl_decrypt('aes128', 'cbc', True)

    def test_email_ssl_decrypt_aes192_cbc_True(self):
        self._test_email_ssl_decrypt('aes192', 'cbc', True)

    def test_email_ssl_decrypt_aes256_cbc_True(self):
        self._test_email_ssl_decrypt('aes256', 'cbc', True)

    def test_email_ssl_decrypt_aes128_ofb_True(self):
        self._test_email_ssl_decrypt('aes128', 'ofb', True)

    def test_email_ssl_decrypt_aes192_ofb_True(self):
        self._test_email_ssl_decrypt('aes192', 'ofb', True)

    def test_email_ssl_decrypt_aes256_ofb_True(self):
        self._test_email_ssl_decrypt('aes256', 'ofb', True)

    def _test_email_ssl_encrypt_smime(self, algo):
        fname = fixture('smime-ssl-encrypted-smime-{}.txt'.format(algo))
        cmd = [
            'openssl', 'smime', '-encrypt', '-'+algo,
            '-in', fixture('smime-unsigned.txt'),
            '-out', fname,
            test_cert.cert1_cert,
        ]
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        assert stderr == b''
        assert stdout == b''

        key = test_cert.CA().key_load(test_cert.cert1_key, '1234')
        with io.open(fname, 'rt', encoding='utf-8') as fh:
            datae = fh.read()
        datad = email.decrypt(datae, key)
        with open(fixture('smime-unsigned.txt'), 'rb') as fh:
            datau = fh.read()

        assert datau == datad.replace(b'\r\n', b'\n')

    def test_email_ssl_encrypt_aes256(self):
        self._test_email_ssl_encrypt_smime('aes256')

    def _test_email_ssl_encrypt_cms(self, mode):
        fname = fixture('smime-ssl-encrypted-cms-{}.txt'.format(mode))
        cmd = [
            'openssl', 'cms', '-encrypt',
            '-recip', test_cert.cert1_cert,
            '-in', fixture('smime-unsigned.txt'),
            '-out', fname,
            '-md', 'sha512'
        ]
        if mode is not None:
            cmd.extend([
                '-keyopt', 'rsa_padding_mode:{}'.format(mode),
            ])
        process = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        assert stderr == b''
        assert stdout == b''

        key = test_cert.CA().key_load(test_cert.cert1_key, '1234')
        with io.open(fname, 'rt', encoding='utf-8') as fh:
            datae = fh.read()
        datad = email.decrypt(datae, key)
        with open(fixture('smime-unsigned.txt'), 'rb') as fh:
            datau = fh.read()

        assert datau == datad.replace(b'\r\n', b'\n')

    def test_email_ssl_encrypt_cms_oaep(self):
        self._test_email_ssl_encrypt_cms('oaep')

    def test_email_ssl_encrypt_cms(self):
        self._test_email_ssl_encrypt_cms(None)



#!/usr/bin/env vpython3

""" A tool which takes in pdf, pfx file, password as input and gives out a corresponding signed pdf

"""
import argparse
import pytz
import re
import sys
import datetime

from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.x509.oid import NameOID
from endesive import pdf

signature_string = lambda organization, date, country : (organization + '\nDATE: '+ date)

def eprint(error):
    print(error, file=sys.stderr)

def load_pfx(file_path, password):
    """ Function to load pkcs12 object from the given password protected pfx file."""

    with open(file_path, 'rb') as fp:
        return pkcs12.load_key_and_certificates(fp.read(), password.encode(), backends.default_backend())

def create_args():
    """Creates CLI arguments for the pdfSigner script."""

    parser = argparse.ArgumentParser(description='Script for digitally signing a pdf')
    parser.add_argument('pfx_certificate', type=str, help='Specify keystore file in .pfx format (Mandatory)')
    parser.add_argument('password', type=str, help=' Specify password for keystore file (mandatory)')
    parser.add_argument('src', type=str,
        help='Specify the source file (.pdf) that needs to be digitally signed. Only 1 file at a time can be signed. (Mandatory) ')
    parser.add_argument('-d', '--dest', type=str,
        help='Specify the destination file where digitally signed content will be stored.When not specified, by default it will '
        'digitally sign the source file.(Mandatory) \n'
        'E.g. Given source file /var/hp/some.pdf will be digitally signed')
    parser.add_argument('-c', '--coords', type=str,
        help='Specify the co-ordinates of where you want the digital signature to be placed on the PDF file page.(Optional)\n'
        'Format: Accepts 4 comma-separated float values (without spaces). E.g. 1,2,3,4 ')
    parser.add_argument('-p', '--page', type=int,
        help='You can specify the page number of PDF file where digital signature(Optional)')

    return parser.parse_args()

def validate_args(args):
    """Validating commandline arguments raises valueError exception with if any command
    line arguments are not valid."""

    IS_PFX = lambda pfx_certificate: re.match( r'^(.[^,]+)(.pfx|.PFX){1}$', pfx_certificate)
    if not IS_PFX(args.pfx_certificate):
        raise ValueError('Not a proper pfx file with .pfx or .PFX extension')
    if args.coords:
        for num in args.coords.split(','):
            if not num.isdigit():
                raise ValueError('Coords are not integers')


OID_NAMES = {
    NameOID.COMMON_NAME: 'CN',
    NameOID.COUNTRY_NAME: 'C',
    NameOID.DOMAIN_COMPONENT: 'DC',
    NameOID.EMAIL_ADDRESS: 'E',
    NameOID.GIVEN_NAME: 'G',
    NameOID.LOCALITY_NAME: 'L',
    NameOID.ORGANIZATION_NAME: 'O',
    NameOID.ORGANIZATIONAL_UNIT_NAME: 'OU',
    NameOID.SURNAME: 'SN'
}
def get_rdns_names(rdns):
    names = {}
    for oid in OID_NAMES:
        names[OID_NAMES[oid]] = ''
    for rdn in rdns:
        for attr in rdn._attributes:
            if attr.oid in OID_NAMES:
                names[OID_NAMES[attr.oid]] = attr.value
    return names

def run():
    args = create_args()

    try:
        validate_args(args)
    except ValueError as e:
        import traceback; traceback.print_exc()
        sys.exit(1)

    try:
        # Load the PKCS12 object from the pfx file
        p12pk, p12pc, p12oc = load_pfx(args.pfx_certificate, args.password)

        names = get_rdns_names(p12pc.subject.rdns)
        timezone = pytz.timezone('Asia/Calcutta')
        #default coords of bottom right corner in a pdf page
        coords = [350, 50, 550, 150]
        if args.coords:
          coords = [int(coord) for coord in args.coords.split(',') if coord]
        page = args.page if args.page else 1
        dest = args.dest if args.dest else args.src
        date = datetime.datetime.utcnow() - datetime.timedelta(hours=12)
        date = date.strftime('%Y%m%d%H%M%S+00\'00\'')
        signature = signature_string(names['CN'], date, names['C'])
        dct = {
            'sigflags': 3,
            'sigpage': page - 1,
            'contact': 'finacctind@endurance.com',
            #'location': subject.C.encode(),
            'location': 'Szczecin',
            'signingdate': date,
            'signingdate': '20201119195200+02\'00\'',
            'reason': 'Signed by endurance',
            'signature': signature,
            'signaturebox': tuple(coords[:4]),
        }

        input_file = args.src
        datau = open(input_file, 'rb').read()
        datas = pdf.cms.sign(datau,
                     dct,
                     p12pk, p12pc, p12oc,
                     'sha256'
                     )

        output_file = input_file.replace(input_file, dest)
        with open(output_file, 'wb') as fp:
          fp.write(datau)
          fp.write(datas)
    except Exception as e:
        import traceback; traceback.print_exc()
        eprint(e)
        sys.exit()

if __name__ == '__main__':
    run()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import os
import stat
import subprocess
import datetime
import base64
import email

from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
from email.mime.base import MIMEBase

from cryptography.hazmat import backends
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import pkcs12

from endesive.hsm import SSHAgentHSM
from endesive import signer


def compose(From, To, Subject, Body, Attachment, signer):
    # create message object instance
    msg = MIMEMultipart(_subtype="signed", micalg="SHA1", protocol="application/pkcs7-signature")

    # setup the parameters of the message
    msg['From'] = From
    msg['To'] = To
    msg['Subject'] = Subject
    msg['Date'] = email.utils.format_datetime(datetime.datetime.now())
    msg.preamble = "This is a multipart message in MIME format."

    env = MIMEMultipart(_subtype='mixed')
    body = MIMEText(Body.decode())
    del body['MIME-Version']
    env.attach(body)

    app = MIMEApplication(open(Attachment, 'rb').read(), _subtype="pdf")
    app.add_header('content-disposition', 'attachment', filename=Attachment)
    env.attach(app)

    msg.attach(env)

    sig = MIMEBase(_maintype='application', _subtype='pkcs7-signature', name="smime.p7s")
    sig.add_header('Content-Disposition', 'attachment', filename='smime.p7s')
    sig.add_header('Content-Transfer-Encoding', 'base64')
    sig.set_payload(signer(env.as_string().encode()))
    del sig['MIME-Version']
    msg.attach(sig)

    return msg, env, sig


def sign(datau, key, cert, othercerts, hashalgo, hsm):
    datau = datau.replace(b'\n', b'\r\n')
    datas = signer.sign(datau, key, cert, othercerts, hashalgo, attrs=True, pss=False, hsm=hsm)
    return base64.encodebytes(datas)


def main():
    # split certificate
    # we need the key as seperate file
    with open('ca/demo2_user1.p12', 'rb') as fp:
        key, cert, othercerts = pkcs12.load_key_and_certificates(fp.read(), b'1234', backends.default_backend())

    agent = SSHAgentHSM(cert)

    # lookup the ssh fingerprint for the certificates public key
    keyid, _ = agent.certificate()
    keyfile = None
    try:
        # is the public key known to the ssh-agent yet?
        agent.key(keyid)
    except ValueError:

        # set file permissions to something ssh-agent accepts
        def perms(path):
            if stat.S_IMODE(os.stat(path).st_mode) & ~stat.S_IRWXU:
                os.chmod(path, (stat.S_IRUSR | stat.S_IWUSR))


        # we have to add the key to the ssh-agent
        # remove the key password, dump in traditional openssl so ssh-agent can add the key
        keyfile = 'ca/demo2_user1.key.nopass.pem'
        with open(keyfile, 'wb') as fp:
            # dump the key
            fp.write(key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption()
            ))
        perms(keyfile)

        # pub file so ssh-add -d can be used
        pubfile = 'ca/demo2_user1.key.nopass.pem.pub'
        with open(pubfile, 'wb') as fp:
            # convert the public key of the certificate to ssh public key format
            fp.write(cert.public_key().public_bytes(
                encoding=serialization.Encoding.OpenSSH,
                format=serialization.PublicFormat.OpenSSH
            ))
        perms(pubfile)

        subprocess.call(["ssh-add", keyfile])

        # reconnect the agent so the key is visible to paramiko
        agent = SSHAgentHSM(cert)


    datau = open('smime-unsigned.txt', 'rb').read()

    msg, env, sig = compose(
        From='root+from@localhost',
        To='root+to@localhost',
        Subject='this is the subject',
        Body=datau,
        Attachment='pdf-acrobat.pdf',
        signer=lambda data: sign(data, None, cert, othercerts, 'sha256', agent)
    )
    datas = msg.as_bytes(unixfrom=True)
    open('smime-signed-hsm.txt', 'wb').write(datas)

    # we added, so we remove the key from ssh-agent
    if keyfile:
        subprocess.call(['ssh-add', '-d', keyfile])

main()



# -*- coding: utf-8 -*-
"""
    Image
    ~~~~~~~~~~~~
    Image annotation class.

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
import sys
import zlib
from io import BytesIO

from PIL import Image as PILImage
from PIL.ImageFile import ImageFile

from ..pdfrw import PdfDict, PdfName, IndirectPdfDict

from .rect import RectAnnotation
from ..config.appearance import set_appearance_state
from ..config.constants import CMYK_MODE
from ..config.constants import GRAYSCALE_ALPHA_MODE
from ..config.constants import GRAYSCALE_MODE
from ..config.constants import PALETTE_MODE
from ..config.constants import RGB_MODE
from ..config.constants import RGBA_MODE
from ..config.constants import SINGLE_CHANNEL_MODE
from ..graphics import ContentStream
from ..graphics import CTM
from ..graphics import Rect
from ..graphics import Restore
from ..graphics import Save
from ..graphics import XObject
from ..util.geometry import matrix_multiply
from ..util.geometry import scale
from ..util.geometry import translate


class Image(RectAnnotation):
    """A basic Image annotation class.

    There is no native image annotation in the PDF spec, but image annotations
    can be easily approximated by drawing an Image XObject in the appearance
    stream.

    Supported formats are PNG, JPEG, and GIF. Only the first frame from multi-
    frame GIFs is used.

    This implementation relies on the python Pillow library to retrieve the
    image's raw sample data, then formats that data as expected by the PDF
    spec.

    Additional work needs to be done. For example:
        - supporting the reading of transparency directly from RGBA images
        - better compression (e.g. using a Predictor value for FlateDecode)
        - supporting DeviceCMYK directly
    """

    subtype = "Square"
    _image_xobject = None  # PdfDict of Image XObject

    def add_additional_resources(self, resources):
        resources[PdfName("XObject")] = IndirectPdfDict(Image=self.image_xobject)

    def add_additional_pdf_object_data(self, obj):
        obj[PdfName("Image")] = self.image_xobject

    @property
    def image_xobject(self):
        if self._image_xobject is None:
            self._image_xobject = self.make_image_xobject(self._appearance.image)
        return self._image_xobject

    @staticmethod
    def make_image_xobject(image):
        """Construct a PdfDict representing the Image XObject, for inserting
        into the AP Resources dict.

        PNGs and GIFs are treated equally - the raw sample values are included
        using PDF's FlateDecode compression format. JPEGs can be included in
        their original form using the DCTDecode filter.

        PNGs with transparency have the alpha channel split out and included as
        an SMask, since PDFs don't natively support transparent PNGs.

        Details about file formats and allowed modes can be found at
        https://pillow.readthedocs.io/en/5.3.x/handbook/image-file-formats.html

        :param str|ImageFile image: Either a str representing the path to the
            image filename, or a PIL.ImageFile.ImageFile object representing
            the image loaded using the PIL library.
        :returns PdfDict: Image XObject
        """
        image = Image.resolve_image(image)
        # PILImage.convert drops the format attribute
        image_format = image.format
        width, height = image.size

        # Normalize images to RGB or grayscale color spaces, and split out the
        # alpha layer into a PDF smask XObject
        image, smask_xobj = Image.convert_to_compatible_image(image, image_format)

        if image_format in ("PNG", "GIF"):
            content = Image.make_compressed_image_content(image)
            filter_type = "FlateDecode"  # TODO use a predictor
        elif image_format == "JPEG":
            content = Image.make_jpeg_image_content(image)
            filter_type = "DCTDecode"
        else:
            raise ValueError(
                "Unsupported image format: {}. Supported formats are "
                "PNG, JPEG, and GIF".format(image.format)
            )

        xobj = IndirectPdfDict(
            stream=content,
            BitsPerComponent=8,
            Filter=PdfName(filter_type),
            ColorSpace=Image._get_color_space_name(image),
            Width=width,
            Height=height,
            Subtype=PdfName("Image"),
            Type=PdfName("XObject"),
        )
        if smask_xobj is not None:
            xobj[PdfName("SMask")] = smask_xobj
        return xobj

    @staticmethod
    def convert_to_compatible_image(image, image_format):
        smask_xobj = None

        if image_format in ("PNG", "GIF"):
            if image.mode in (RGBA_MODE, GRAYSCALE_ALPHA_MODE):
                smask_xobj = Image.get_png_smask(image)

            # 'P' and 'RGBA' images can just be converted to 'RGB' mode, since
            # alpha layer is preserved in the smask.
            if image.mode in (RGBA_MODE, PALETTE_MODE):
                image = image.convert(RGB_MODE)

            if image.mode == GRAYSCALE_ALPHA_MODE:
                image = image.convert(GRAYSCALE_MODE)

        if image.mode == CMYK_MODE:
            # The DeviceCMYK PDF color space has some weird properties. In a
            # future release we can debug these and be smart about it but this
            # is an easy workaround.
            image = image.convert(RGB_MODE)

        return image, smask_xobj

    @staticmethod
    def get_png_smask(image):
        width, height = image.size
        smask = Image.make_compressed_image_content(image.getchannel("A"))
        smask_xobj = IndirectPdfDict(
            stream=smask,
            Width=width,
            Height=height,
            BitsPerComponent=8,
            Filter=PdfName("FlateDecode"),
            ColorSpace=PdfName("DeviceGray"),
            Subtype=PdfName("Image"),
            Type=PdfName("XObject"),
        )
        return smask_xobj

    @staticmethod
    def resolve_image(image_or_filename):
        if isinstance(image_or_filename, str):
            return PILImage.open(image_or_filename)
        elif isinstance(image_or_filename, ImageFile):
            return image_or_filename
        raise ValueError(
            "Invalid image format: {}".format(image_or_filename.__class__.__name__)
        )

    @staticmethod
    def _get_color_space_name(image):
        if image.mode == RGB_MODE:
            return PdfName("DeviceRGB")
        elif image.mode in (GRAYSCALE_MODE, SINGLE_CHANNEL_MODE):
            return PdfName("DeviceGray")
        raise ValueError("Image color space not yet supported")

    @staticmethod
    def make_compressed_image_content(image):
        compressed = zlib.compress(Image.get_raw_image_bytes(image))
        return compressed

    @staticmethod
    def make_jpeg_image_content(image):
        file_obj = BytesIO()
        # This is the only way to preserve the JPEG encoding. It also seems to
        # recompress the data, so that the raw bytes of this differ from the
        # raw bytes of the original file. It'll probably be better to provide a
        # special wrapper around PILImage that preserves the original bytes so
        # we can just use those for JPEGs. TODO.
        image.save(file_obj, format="JPEG")
        return file_obj.getvalue()

    @staticmethod
    def get_decoded_bytes(content):
        # Right now, pdfrw needs strings, not bytes like you'd expect in py3,
        # for binary stream objects. This might change in future versions.
        if sys.version_info.major < 3:
            return content
        return content.decode("Latin-1")

    @staticmethod
    def get_raw_image_bytes(image):
        if image.mode in (GRAYSCALE_MODE, SINGLE_CHANNEL_MODE):
            # If this is grayscale or single-channel, we can avoid dealing with
            # the nested tuples in multi-channel images. This bytes/bytearray
            # wrapped approach is the only way that works in both py2 and py3.
            return bytes(bytearray(image.getdata()))

        elif image.mode == RGB_MODE:
            raw_image_data = list(image.getdata())
            array = bytearray()
            for rgb in raw_image_data:
                array.extend(rgb)
            return bytes(array)

        raise ValueError("Image color space not yet supported")

    @staticmethod
    def get_ctm(x1, y1, x2, y2):
        """Get the scaled and translated CTM for an image to be placed in the
        bounding box defined by [x1, y1, x2, y2].
        """
        return matrix_multiply(translate(x1, y1), scale(x2 - x1, y2 - y1))

    def make_appearance_stream(self):
        A = self._appearance
        L = self._location

        stream = ContentStream([Save()])
        set_appearance_state(stream, A)
        stream.extend(
            [
                Rect(L.x1, L.y1, L.x2 - L.x1, L.y2 - L.y1),
                CTM(self.get_ctm(L.x1, L.y1, L.x2, L.y2)),
                XObject("Image"),
                Restore(),
            ]
        )
        return stream



# -*- coding: utf-8 -*-
"""
    Constants
    ~~~~~~~~~

    :copyright: Copyright 2019 Autodesk, Inc.
    :license: MIT, see LICENSE for details.
"""
TEXT_ALIGN_LEFT = 'left'
TEXT_ALIGN_CENTER = 'center'
TEXT_ALIGN_RIGHT = 'right'
ALLOWED_ALIGNS = (
    TEXT_ALIGN_LEFT,
    TEXT_ALIGN_RIGHT,
    TEXT_ALIGN_CENTER,
)

TEXT_BASELINE_TOP = 'top'
TEXT_BASELINE_MIDDLE = 'middle'
TEXT_BASELINE_BOTTOM = 'bottom'
ALLOWED_BASELINES = (
    TEXT_BASELINE_TOP,
    TEXT_BASELINE_MIDDLE,
    TEXT_BASELINE_BOTTOM,
)

DEFAULT_CONTENT = ''
DEFAULT_FONT_SIZE = 12
DEFAULT_LINE_SPACING = 1.2

DEFAULT_STROKE_WIDTH = 1
DEFAULT_BORDER_STYLE = 'S'

LINE_CAP_BUTT = 0
LINE_CAP_ROUND = 1
LINE_CAP_SQUARE = 2
ALLOWED_LINE_CAPS = (
    LINE_CAP_BUTT,
    LINE_CAP_ROUND,
    LINE_CAP_SQUARE,
)

LINE_JOIN_MITER = 0
LINE_JOIN_ROUND = 1
LINE_JOIN_BEVEL = 2
ALLOWED_LINE_JOINS = (
    LINE_JOIN_MITER,
    LINE_JOIN_ROUND,
    LINE_JOIN_BEVEL,
)

BLACK = (0, 0, 0)
TRANSPARENT = tuple()

GRAPHICS_STATE_NAME = 'PdfAnnotatorGS'
DEFAULT_BASE_FONT = 'Helvetica'
PDF_ANNOTATOR_FONT = 'PDFANNOTATORFONT1'

CMYK_MODE = 'CMYK'
RGB_MODE = 'RGB'
RGBA_MODE = 'RGBA'
PALETTE_MODE = 'P'
GRAYSCALE_MODE = 'L'
GRAYSCALE_ALPHA_MODE = 'LA'
SINGLE_CHANNEL_MODE = '1'



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import base64
import json
from asn1crypto import pem as asn1pem
from endesive import hsm

import os
import sysconfig

os.environ["SOFTHSM2_CONF"] = "softhsm2.conf"
if not os.path.exists(os.path.join(os.getcwd(), "softhsm2.conf")):
    open("softhsm2.conf", "wt").write(
        """\
log.level = DEBUG
directories.tokendir = %s/softhsm2/
objectstore.backend = file
slots.removable = false
"""
        % os.getcwd()
    )
if not os.path.exists(os.path.join(os.getcwd(), "softhsm2")):
    os.mkdir(os.path.join(os.getcwd(), "softhsm2"))

#
#!/bin/bash
# SOFTHSM2_CONF=softhsm2.conf
# softhsm2-util --label "endesive" --slot 1 --init-token --pin secret1 --so-pin secret2
# softhsm2-util --show-slots
#

if sys.platform == "win32":
    dllpath = r"W:\binw\SoftHSM2\lib\softhsm2-x64.dll"
else:
    dllpath = os.path.join(sysconfig.get_config_var('LIBDIR'), "softhsm/libsofthsm2.so")

import PyKCS11 as PK11


class Signer(hsm.HSM):
    def certificate(self):
        self.login("endesieve", "secret1")
        keyid = bytes((0x66, 0x66, 0x90))
        try:
            pk11objects = self.session.findObjects(
                [(PK11.CKA_CLASS, PK11.CKO_CERTIFICATE)]
            )
            all_attributes = [
                # PK11.CKA_SUBJECT,
                PK11.CKA_VALUE,
                # PK11.CKA_ISSUER,
                # PK11.CKA_CERTIFICATE_CATEGORY,
                # PK11.CKA_END_DATE,
                PK11.CKA_ID,
            ]

            for pk11object in pk11objects:
                try:
                    attributes = self.session.getAttributeValue(
                        pk11object, all_attributes
                    )
                except PK11.PyKCS11Error as e:
                    continue

                attrDict = dict(list(zip(all_attributes, attributes)))
                cert = bytes(attrDict[PK11.CKA_VALUE])
                if keyid == bytes(attrDict[PK11.CKA_ID]):
                    return keyid, cert
        finally:
            self.logout()
        return None, None

    def sign(self, keyid, data, mech):
        self.login("endesieve", "secret1")
        try:
            privKey = self.session.findObjects(
                [(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, keyid)]
            )[0]
            mech = getattr(PK11, "CKM_%s_RSA_PKCS" % mech.upper())
            sig = self.session.sign(privKey, data, PK11.Mechanism(mech, None))
            return bytes(sig)
        finally:
            self.logout()


def main():
    pdfname = 'pdf.pdf'
    if len (sys.argv) > 1:
        pdfname = sys.argv[1]
    config = open(pdfname + ".json", "rt").read()
    config = json.loads(config)

    tosign = base64.decodebytes(config['tosign'].encode('ascii'))

    clshsm = Signer(dllpath)
    keyid, cert = clshsm.certificate()
    signed_bytes = clshsm.sign(keyid, tosign, "sha256")

    config['signed_bytes'] = b"".join(base64.encodebytes(signed_bytes).split()).decode('ascii')
    config['certificate'] = asn1pem.armor("CERTIFICATE", cert).decode('ascii')
    json.dump(config, open(pdfname + ".json", "wt"), indent=4)


main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import datetime
from oscrypto import keys
from endesive import pdf

#import logging
#logging.basicConfig(level=logging.DEBUG)

def main():
    date = datetime.datetime.utcnow() - datetime.timedelta(hours=12)
    date = date.strftime('%Y%m%d%H%M%S+00\'00\'')
    dct = {
        'sigflags': 3,
        # 'sigpage': 0,
        'sigbutton': True,
        'signature_img': 'signature_test.png',
        'contact': 'mak@trisoft.com.pl',
        'location': 'Szczecin',
        'signingdate': date.encode(),
        'reason': 'Dokument podpisany cyfrowo',
        'signature': 'Dokument podpisany cyfrowo',
        'signaturebox': (470, 0, 570, 100),
    }
    with open('ca/demo2_user1.p12', 'rb') as f:
        key, cert, certe = keys.parse_pkcs12(f.read(), b'1234')
    fname = 'pdf.pdf'
    if len (sys.argv) > 1:
        fname = sys.argv[1]
    datau = open(fname, 'rb').read()
    datas = pdf.cms.sign(datau, dct,
        key,
        cert,
        certe,
        'sha256'
    )
    fname = fname.replace('.pdf', '-signed-cms-oscrypto.pdf')
    with open(fname, 'wb') as fp:
        fp.write(datau)
        fp.write(datas)


main()



#!/usr/bin/env vpython3
import os

jars = [
'java/debugger-app-2.0.14.jar',
'java/fontbox-2.0.14.jar',
'java/java.activation-1.1.1.jar',
'java/javax.xml.bind-2.2.3.jar',
'java/pdfbox-2.0.14.jar',
'java/pdfbox-app-2.0.14.jar',
'java/pdfbox-debugger-2.0.14.jar',
'java/pdfbox-showsignature.jar',
'java/pdfbox-tools-2.0.14.jar',
'java/preflight-2.0.14.jar',
'java/preflight-app-2.0.14.jar',
'java/xades4j-1.5.1.jar',
'java/xmpbox-2.0.14.jar',
]
os.environ['CLASSPATH'] = ':'.join(jars)
os.environ['JAVA_HOME'] = '/usr/lib/jvm/default-java'

import jnius



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import datetime
from certum import dllpath
from cryptography import x509
from cryptography.hazmat import backends
import PyKCS11 as PK11
from endesive import pdf, hsm


class Signer(hsm.HSM):
    def certificate(self):
        self.login("profil bezpieczny", "9593")
        keyid = [0x5e, 0x9a, 0x33, 0x44, 0x8b, 0xc3, 0xa1, 0x35, 0x33, 0xc7, 0xc2, 0x02, 0xf6, 0x9b, 0xde, 0x55, 0xfe, 0x83, 0x7b, 0xde]
        #keyid = [0x3f, 0xa6, 0x63, 0xdb, 0x75, 0x97, 0x5d, 0xa6, 0xb0, 0x32, 0xef, 0x2d, 0xdc, 0xc4, 0x8d, 0xe8]
        keyid = bytes(keyid)
        try:
            pk11objects = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_CERTIFICATE)])
            all_attributes = [
                #PK11.CKA_SUBJECT,
                PK11.CKA_VALUE,
                #PK11.CKA_ISSUER,
                #PK11.CKA_CERTIFICATE_CATEGORY,
                #PK11.CKA_END_DATE,
                PK11.CKA_ID,
            ]

            for pk11object in pk11objects:
                try:
                    attributes = self.session.getAttributeValue(pk11object, all_attributes)
                except PK11.PyKCS11Error as e:
                    continue

                attrDict = dict(list(zip(all_attributes, attributes)))
                cert = bytes(attrDict[PK11.CKA_VALUE])
                if keyid == bytes(attrDict[PK11.CKA_ID]):
                    return keyid, cert
        finally:
            self.logout()
        return None, None

    def sign(self, keyid, data, mech):
        self.login("profil bezpieczny", "9593")
        try:
            privKey = self.session.findObjects([(PK11.CKA_CLASS, PK11.CKO_PRIVATE_KEY), (PK11.CKA_ID, keyid)])[0]
            mech = getattr(PK11, 'CKM_%s_RSA_PKCS' % mech.upper())
            sig = self.session.sign(privKey, data, PK11.Mechanism(mech, None))
            return bytes(sig)
        finally:
            self.logout()

def main():
    tspurl = "http://time.certum.pl"
    tspurl = "http://public-qlts.certum.pl/qts-17"
    date = datetime.datetime.utcnow() - datetime.timedelta(hours=12)
    date = date.strftime('%Y%m%d%H%M%S+00\'00\'')
    dct = {
        'sigflags': 3,
        'sigpage': 0,
        'sigbutton': True,
        'contact': 'mak@trisoft.com.pl',
        'location': 'Szczecin',
        'signingdate': date.encode(),
        'reason': 'Dokument podpisany cyfrowo',
        'signature': 'Dokument podpisany cyfrowo',
        'signaturebox': (0, 0, 100, 100),
    }

    ocspurl = 'https://ocsp.certum.pl/'
    ocspissuer = open('CertumDigitalIdentificationCASHA2.crt', 'rb').read()
    ocspissuer = x509.load_pem_x509_certificate(ocspissuer, backends.default_backend())

    clshsm = Signer(dllpath)
    fname = 'pdf.pdf'
    if len (sys.argv) > 1:
        fname = sys.argv[1]
    datau = open(fname, 'rb').read()
    datas = pdf.cms.sign(datau, dct,
        None, None,
        [],
        'sha256',
        clshsm,
        tspurl,
        ocspurl=ocspurl,
        ocspissuer=ocspissuer
    )
    fname = fname.replace('.pdf', '-signed-cms-hsm-certum.pdf')
    with open(fname, 'wb') as fp:
        fp.write(datau)
        fp.write(datas)


main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import datetime
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12

from endesive.pdf import cms

# from endesive.pdf import cmsn as cms

# import logging
# logging.basicConfig(level=logging.DEBUG)


def main():
    dct = {
        "aligned": 0,
        "sigflags": 3,
        "sigflagsft": 132,
        "sigpage": 0,
        "sigfield": "Signature1",
        "auto_sigfield": True,
        "password": "1234",
    }
    fname = "pdf.pdf"
    if len(sys.argv) > 1:
        fname = sys.argv[1]
    datau = open(fname, "rb").read()
    datas = cms.timestamp(
        datau,      # PDF data
        dct,        # config
        "sha256",   # hash
        'https://freetsa.org/tsr' # Timestamp server URL
        # { # Timestamp server credentials
        #     'username': 'user',
        #     'password': 'hunter2'
        #     }, 
        # {}, Timestamp server options 
        )
    fname = fname.replace(".pdf", "-timestamped-cms.pdf")
    with open(fname, "wb") as fp:
        fp.write(datau)
        fp.write(datas)


main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
from endesive import pdf


def main():
    trusted_cert_pems = (
        # demo ca chain
        open("ca/demo2_ca.crt.pem", "rb").read(),
        # demo hsm ca chain
        open("cert-hsm-ca.pem", "rb").read(),
    )
    for fname in (
        "test-PDFXRef-signed-cms.pdf",
        "test-PDFXRefStream-signed-cms.pdf",
        "test-SHA256_RSA-signed-cms.pdf",
        "pdf-acrobat.pdf",
        "pdf-signed-cms-hsm-certum.pdf",
        "pdf-signed-cms-hsm.pdf",
        "pdf-signed-cms-m32-unizeto.pdf",
        "pdf-signed-cms-m32-actalis.pdf",
        "pdf-signed-cms-oscrypto.pdf",
        "pdf-signed-cms-pfx.pdf",
        "pdf-signed-cms.pdf",
        "pdf-signed-fpdf.pdf",
        "pdf-signed-java.pdf",
        "pdf-signed-pypdf.pdf",
        "pdf-encrypted-signed-java.pdf",
        "pdf-encrypted-signed-pypdf.pdf",
        "pdf-link-signed-java.pdf",
        "pdf-link-signed-pypdf.pdf",
    ):
        print("*" * 20, fname)
        try:
            data = open(fname, "rb").read()
        except:
            continue
        no = 0
        try:
            for (hashok, signatureok, certok) in pdf.verify(
                data, trusted_cert_pems, "/etc/ssl/certs"
            ):
                print("*" * 10, "signature no:", no)
                print("signature ok?", signatureok)
                print("hash ok?", hashok)
                print("cert ok?", certok)
        except Exception as exc:
            print(exc)

main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import sys
import datetime
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12

from endesive.pdf import cms

# from endesive.pdf import cmsn as cms

# import logging
# logging.basicConfig(level=logging.DEBUG)


def main():
    date = datetime.datetime.utcnow() - datetime.timedelta(hours=12)
    date = date.strftime("D:%Y%m%d%H%M%S+00'00'")
    dct = {
        "aligned": 0,
        "sigflags": 3,
        "sigflagsft": 132,
        "sigpage": 0,
        #"auto_sigfield": False,
        #"sigandcertify": False,
        #"signaturebox": (0, 0, 590, 155),
        "signform": True,
        "sigfield": "Signature",
        #             Text will be in the default font
        "signature": 'Signed field!',

        # default configuration for the text appearance
        "text": {
            'wraptext': True,
            'fontsize': 12,
            'textalign': 'left',
            'linespacing': 1.2,
            },

        "contact": "mak@trisoft.com.pl",
        "location": "Szczecin",
        "signingdate": date,
        "reason": "Dokument podpisany cyfrowo aą cć eę lł nń oó sś zż zź",
        "password": "1234",
    }
    with open("../ca/demo2_user1.p12", "rb") as fp:
        p12 = pkcs12.load_key_and_certificates(
            fp.read(), b"1234", backends.default_backend()
        )
    fname = "../pdf_forms/blank_form.pdf"
    if len(sys.argv) > 1:
        fname = sys.argv[1]
    datau = open(fname, "rb").read()
    datas = cms.sign(datau, dct, p12[0], p12[1], p12[2], "sha256")
    fname = fname.replace(".pdf", "-signature.pdf")
    with open(fname, "wb") as fp:
        fp.write(datau)
        fp.write(datas)


main()



#!/usr/bin/env vpython3
# coding: utf-8

import os
import sys
import sysconfig

os.environ['SOFTHSM2_CONF'] = 'softhsm2.conf'
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2.conf')):
    open('softhsm2.conf', 'wt').write('''\
log.level = DEBUG
directories.tokendir = %s/softhsm2/
objectstore.backend = file
slots.removable = false
''' % os.getcwd())
if not os.path.exists(os.path.join(os.getcwd(), 'softhsm2')):
    os.mkdir(os.path.join(os.getcwd(), 'softhsm2'))

#
#!/bin/bash
#SOFTHSM2_CONF=softhsm2.conf
#softhsm2-util --label "endesive" --slot 1 --init-token --pin secret1 --so-pin secret2
#softhsm2-util --show-slots
#

if sys.platform == 'win32':
    dllpath = r'W:\binw\SoftHSM2\lib\softhsm2-x64.dll'
else:
    dllpath = os.path.join(sysconfig.get_config_var('LIBDIR'), "softhsm/libsofthsm2.so")

from endesive import hsm, pdf
import PyKCS11 as PK11
from asn1crypto import pem as asn1pem

'''
Create two certificates:
1. self signed CA certificate with serial equal to HSM keyID=0x01
2. USER 1 certificate with serial equal to HSM keyID=0x666690
'''
class HSM(hsm.HSM):
    def main(self):
        cakeyID = bytes((0x1,))
        ca_cert_pem = asn1pem.armor('CERTIFICATE', self.cert_load(cakeyID))
        trusted_cert_pems = (ca_cert_pem,)
        for fname in (
            'pdf-signed-cms-hsm.pdf',
            'pdf-signed-cms-hsm-signed-cms-hsm.pdf',
        ):
            print('*' * 20, fname)
            try:
                data = open(fname, 'rb').read()
            except:
                continue
            results = pdf.verify(data, trusted_cert_pems)
            for i in range(len(results)):
                print('*'*10, 'signature #{}'.format(i+1))
                (hashok, signatureok, certok) = results[i]
                print('signature ok?', signatureok)
                print('hash ok?', hashok)
                print('cert ok?', certok)

def main():
    cls = HSM(dllpath)
    cls.login("endesieve", "secret1")
    try:
        cls.main()
    finally:
        cls.logout()
main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import io
from cryptography.hazmat import backends
from cryptography.hazmat.primitives.serialization import pkcs12
from endesive import email


def main():
    with open('ca/demo2_user1.p12', 'rb') as fp:
        p12 = pkcs12.load_key_and_certificates(fp.read(), b'1234', backends.default_backend())
    for fname in (
        'smime-ssl-encrypted.txt',
        'smime-ssl-oaep-encrypted.txt',
        'smime-encrypted.txt',
        'smime-encrypted-oaep.txt',
    ):
        print('*' * 20, fname)
        try:
            datae = io.open(fname, 'rt', encoding='utf-8').read()
        except:
            print('no such file')
            continue
        datad = email.decrypt(datae, p12[0])
        datad = datad.decode('utf-8')
        io.open(fname.replace('encrypted', 'decrypted'), 'wt', encoding='utf-8').write(datad)


main()



#!/usr/bin/env vpython3
# *-* coding: utf-8 *-*
import io


def main():
    io.open('smime-unsigned.txt', 'wt', encoding='utf-8').write('''\
Witam,

Fantastycznie, że zechciałeś popatrzeć na tę bibliotekę.
Mam nadzieję, że będzie dla Ciebie użyteczna.

Pozdrawiam,
Grzegorz Makarewicz
'''
                                                                )


main()
